# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


import os
import time
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Dict, Literal, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
from ray import train
from torch.utils.data import Dataset, TensorDataset
from tqdm import tqdm

from itwinai.distributed import suppress_workers_print
from itwinai.loggers import EpochTimeTracker, Logger
from itwinai.torch.config import TrainingConfiguration
from itwinai.torch.distributed import DeepSpeedStrategy, RayDDPStrategy, RayDeepSpeedStrategy
from itwinai.torch.monitoring.monitoring import measure_gpu_utilization
from itwinai.torch.profiling.profiler import profile_torch_trainer
from itwinai.torch.trainer import RayTorchTrainer, TorchTrainer
from src.model import Decoder, Decoder_2d_deep, GeneratorResNet, UNet
from src.utils import init_weights


class VirgoTrainingConfiguration(TrainingConfiguration):
    """Virgo TrainingConfiguration"""

    #: Whether to save best model on validation dataset. Defaults to True.
    save_best: bool = True
    #: Loss function. Defaults to "l1".
    loss: Literal["l1", "l2"] = "l1"
    #: Generator to train. Defaults to "unet".
    generator: Literal["simple", "deep", "resnet", "unet"] = "unet"


class NoiseGeneratorTrainer(TorchTrainer):
    def __init__(
        self,
        num_epochs: int = 2,
        config: Dict | TrainingConfiguration | None = None,
        strategy: Literal["ddp", "deepspeed", "horovod"] | None = "ddp",
        checkpoint_path: str = "checkpoints/epoch_{}.pth",
        logger: Logger | None = None,
        random_seed: int | None = None,
        name: str | None = None,
        validation_every: int = 0,
        **kwargs,
    ) -> None:
        super().__init__(
            epochs=num_epochs,
            config=config,
            strategy=strategy,
            logger=logger,
            random_seed=random_seed,
            name=name,
            validation_every=validation_every,
            **kwargs,
        )
        self.save_parameters(**self.locals2params(locals()))
        # Global training configuration

        if isinstance(config, dict):
            config = VirgoTrainingConfiguration(**config)
        self.config = config
        self.num_epochs = num_epochs
        self.checkpoints_location = checkpoint_path
        os.makedirs(os.path.dirname(checkpoint_path), exist_ok=True)

    def create_model_loss_optimizer(self) -> None:
        # Select generator

        generator = self.config.generator.lower()
        scaling = 0.02
        if generator == "simple":
            self.model = Decoder(3, norm=False)
        elif generator == "deep":
            self.model = Decoder_2d_deep(3)
        elif generator == "resnet":
            self.model = GeneratorResNet(3, 12, 1)
            scaling = 0.01
        elif generator == "unet":
            self.model = UNet(input_channels=3, output_channels=1, norm=False)
        else:
            raise ValueError("Unrecognized generator type! Got", generator)

        init_weights(self.model, "normal", scaling=scaling)

        loss = self.config.loss.lower()
        if loss == "l1":
            self.loss = nn.L1Loss()
        elif loss == "l2":
            self.loss = nn.MSELoss()
        else:
            raise ValueError("Unrecognized loss type! Got", loss)

        # Optimizer
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.config.optim_lr)

        # IMPORTANT: model, optimizer, and scheduler need to be distributed

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        else:
            distribute_kwargs = {}

        # Distributed model, optimizer, and scheduler
        self.model, self.optimizer, _ = self.strategy.distributed(
            self.model, self.optimizer, **distribute_kwargs
        )

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> None:
        """Override the create_dataloaders function to use the custom_collate function."""
        # This is the case if a small dataset is used in-memory
        # - we can use the default collate_fn function
        if isinstance(train_dataset, TensorDataset):
            return super().create_dataloaders(
                train_dataset=train_dataset,
                validation_dataset=validation_dataset,
                test_dataset=test_dataset,
            )
        else:
            # If we are using a custom dataset for the large dataset,
            # we need to overwrite the collate_fn function
            self.train_dataloader = self.strategy.create_dataloader(
                dataset=train_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers_dataloader,
                pin_memory=self.config.pin_gpu_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_train,
                collate_fn=self.custom_collate,
            )
            if validation_dataset is not None:
                self.validation_dataloader = self.strategy.create_dataloader(
                    dataset=validation_dataset,
                    batch_size=self.config.batch_size,
                    num_workers=self.config.num_workers_dataloader,
                    pin_memory=self.config.pin_gpu_memory,
                    generator=self.torch_rng,
                    shuffle=self.config.shuffle_validation,
                    collate_fn=self.custom_collate,
                )
            if test_dataset is not None:
                self.test_dataloader = self.strategy.create_dataloader(
                    dataset=test_dataset,
                    batch_size=self.config.batch_size,
                    num_workers=self.config.num_workers_dataloader,
                    pin_memory=self.config.pin_gpu_memory,
                    generator=self.torch_rng,
                    shuffle=self.config.shuffle_test,
                    collate_fn=self.custom_collate,
                )

    def custom_collate(self, batch):
        """
        Custom collate function to concatenate input tensors along their first dimension.
        """
        # Some batches contain None values, if any files from the dataset did not match the
        # criteria (i.e. three auxilliary channels)
        batch = [x for x in batch if x is not None]

        return torch.cat(batch)

    @suppress_workers_print
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Optional[Dataset] = None,
        test_dataset: Optional[Dataset] = None,
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        return super().execute(train_dataset, validation_dataset, test_dataset)

    @profile_torch_trainer
    @measure_gpu_utilization
    def train(self):
        # Start the timer for profiling
        #
        st = timer()
        # uncomment all lines relative to accuracy if you want to measure
        # IOU between generated and real spectrograms.
        # Note that it significantly slows down the whole process
        # it also might not work as the function has not been fully
        # implemented yet
        epoch_time_logger: EpochTimeTracker | None = None
        if self.strategy.is_main_worker and self.strategy.is_distributed:
            print("TIMER: broadcast:", timer() - st, "s")
            print("\nDEBUG: start training")
            print("--------------------------------------------------------")
            # nnod = os.environ.get('SLURM_NNODES', 'unk')
            # s_name = f"{os.environ.get('DIST_MODE', 'unk')}-torch"
            # save_path

            num_nodes = int(os.environ.get("SLURM_NNODES", 1))
            epoch_time_output_dir = Path(f"scalability-metrics/{self.run_id}/epoch-time")
            epoch_time_file_name = f"epochtime_{self.strategy.name}_{num_nodes}N.csv"
            epoch_time_output_path = epoch_time_output_dir / epoch_time_file_name
            epoch_time_logger = EpochTimeTracker(
                strategy_name=self.strategy.name,
                save_path=epoch_time_output_path,
                num_nodes=num_nodes,
                should_log=self.measure_epoch_time
            )
        loss_plot = []
        val_loss_plot = []
        acc_plot = []
        val_acc_plot = []
        best_val_loss = float("inf")

        for epoch in tqdm(range(self.num_epochs)):
            lt = timer()
            # itwinai - IMPORTANT: set current epoch ID
            self.set_epoch(epoch)
            t_list = []
            st = time.time()
            epoch_loss = []
            # epoch_acc = []
            for i, batch in enumerate(self.train_dataloader):
                t = timer()
                # The TensorDataset returns batches as lists of length 1
                if isinstance(batch, list):
                    batch = batch[0]
                # batch= transform(batch)
                target = batch[:, 0].unsqueeze(1).to(self.device)
                # print(f'TARGET ON DEVICE: {target.get_device()}')
                target = target.float()
                input = batch[:, 1:].to(self.device)
                # print(f'INPUT ON DEVICE: {input.get_device()}')

                self.optimizer.zero_grad()
                generated = self.model(input.float())
                # generated=normalize_(generated,1)
                loss = self.loss(generated, target)
                loss.backward()
                self.optimizer.step()
                epoch_loss.append(loss.detach().cpu().numpy())
                t_list.append(timer() - t)
                # itwinai - log loss as metric
                self.log(
                    loss.detach().cpu().numpy(),
                    "epoch_loss_batch",
                    kind="metric",
                    step=epoch * len(self.train_dataloader) + i,
                    batch_idx=i,
                )
                # acc=accuracy(generated.detach().cpu().numpy(),target.detach().cpu().numpy(),20)
                # epoch_acc.append(acc)
            if self.strategy.is_main_worker:
                print("TIMER: train time", sum(t_list) / len(t_list), "s")
            val_loss = []
            # val_acc = []
            for i, batch in enumerate(self.validation_dataloader):
                # batch= transform(batch)
                if isinstance(batch, list):
                    batch = batch[0]
                target = batch[:, 0].unsqueeze(1).to(self.device)
                target = target.float()
                input = batch[:, 1:].to(self.device)
                with torch.no_grad():
                    generated = self.model(input.float())
                    # generated=normalize_(generated,1)
                    loss = self.loss(generated, target)
                val_loss.append(loss.detach().cpu().numpy())
                # itwinai -log loss as metric
                self.log(
                    loss.detach().cpu().numpy(),
                    "val_loss_batch",
                    kind="metric",
                    step=epoch * len(self.validation_dataloader) + i,
                    batch_idx=i,
                )
                # acc=accuracy(generated.detach().cpu().numpy(),target.detach().cpu().numpy(),20)
                # val_acc.append(acc)
            loss_plot.append(np.mean(epoch_loss))
            val_loss_plot.append(np.mean(val_loss))
            # acc_plot.append(np.mean(epoch_acc))
            # val_acc_plot.append(np.mean(val_acc))

            # itwinai - Log metrics/losses
            self.log(np.mean(epoch_loss), "epoch_loss", kind="metric", step=epoch)
            self.log(np.mean(val_loss), "val_loss", kind="metric", step=epoch)
            # self.log(np.mean(epoch_acc), 'epoch_acc',
            #          kind='metric', step=epoch)
            # self.log(np.mean(val_acc), 'val_acc',
            #          kind='metric', step=epoch)

            # print('epoch: {} loss: {} val loss: {} accuracy: {} val
            # accuracy: {}'.format(epoch,loss_plot[-1],val_loss_plot[-1],
            # acc_plot[-1],val_acc_plot[-1]))
            et = time.time()
            # itwinai - print() in a multi-worker context (distributed)
            if self.strategy.is_main_worker:
                print(
                    "epoch: {} loss: {} val loss: {} time:{}s".format(
                        epoch, loss_plot[-1], val_loss_plot[-1], et - st
                    )
                )

            # Save checkpoint every #validation_every epochs
            if self.validation_every and epoch % self.validation_every == 0:
                # uncomment the following if you want to save checkpoint every
                # 100 epochs regardless of the performance of the model
                # checkpoint = {
                #     'epoch': epoch,
                #     'model_state_dict': generator.state_dict(),
                #     'optim_state_dict': optimizer.state_dict(),
                #     'loss': loss_plot[-1],
                #     'val_loss': val_loss_plot[-1],
                # }
                # if self.strategy.is_main_worker:
                #     # Save only in the main worker
                #     checkpoint_filename = checkpoint_path.format(epoch)
                #     torch.save(checkpoint, checkpoint_filename)

                # Average loss among all workers
                # itwinai - gather local loss from all the workers
                worker_val_losses = self.strategy.gather_obj(val_loss_plot[-1])
                if self.strategy.is_main_worker:
                    # Save only in the main worker

                    # avg_loss has a meaning only in the main worker
                    avg_loss = np.mean(worker_val_losses)

                    # instead of val_loss and best_val loss we should
                    # use accuracy!!!
                    if self.config.save_best and avg_loss < best_val_loss:
                        # create checkpoint
                        checkpoint = {
                            "epoch": epoch,
                            "model_state_dict": self.model.state_dict(),
                            "optim_state_dict": self.optimizer.state_dict(),
                            "loss": loss_plot[-1],
                            "val_loss": val_loss_plot[-1],
                        }

                        # save checkpoint only if it is better than
                        # the previous ones
                        checkpoint_filename = self.checkpoints_location.format(epoch)
                        torch.save(checkpoint, checkpoint_filename)
                        # itwinai - log checkpoint as artifact
                        self.log(
                            checkpoint_filename,
                            os.path.basename(checkpoint_filename),
                            kind="artifact",
                        )

                        # update best model
                        best_val_loss = val_loss_plot[-1]
                        best_checkpoint_filename = self.checkpoints_location.format("best")
                        torch.save(checkpoint, best_checkpoint_filename)
                        # itwinai - log checkpoint as artifact
                        self.log(
                            best_checkpoint_filename,
                            os.path.basename(best_checkpoint_filename),
                            kind="artifact",
                        )
            # return (loss_plot, val_loss_plot,
            # acc_plot, val_acc_plot ,acc_plot, val_acc_plot)
            if self.strategy.is_main_worker and self.strategy.is_distributed:
                print("TIMER: epoch time:", timer() - lt, "s")
                assert epoch_time_logger is not None
                epoch_time_logger.add_epoch_time(epoch - 1, timer() - lt)

            # Report training metrics of last epoch to Ray
            train.report({"loss": np.mean(val_loss)})

        return loss_plot, val_loss_plot, acc_plot, val_acc_plot


class RayNoiseGeneratorTrainer(RayTorchTrainer):
    def __init__(
        self,
        config: Dict,
        strategy: Optional[Literal["ddp", "deepspeed"]] = "ddp",
        name: Optional[str] = None,
        logger: Optional[Logger] = None,
        random_seed: int = 1234,
    ) -> None:
        super().__init__(
            config=config, strategy=strategy, name=name, logger=logger, random_seed=random_seed
        )

    def create_model_loss_optimizer(self) -> None:
        # Select generator
        generator = self.training_config.generator
        scaling = 0.02
        if generator == "simple":
            self.model = Decoder(3, norm=False)
        elif generator == "deep":
            self.model = Decoder_2d_deep(3)
        elif generator == "resnet":
            self.model = GeneratorResNet(3, 12, 1)
            scaling = 0.01
        elif generator == "unet":
            self.model = UNet(input_channels=3, output_channels=1, norm=False)
        else:
            raise ValueError("Unrecognized generator type! Got", generator)

        init_weights(self.model, "normal", scaling=scaling)

        # Select loss
        loss = self.training_config.loss
        if loss == "l1":
            self.loss = nn.L1Loss()
        elif loss == "l2":
            self.loss = nn.MSELoss()
        else:
            raise ValueError("Unrecognized loss type! Got", loss)

        # Optimizer
        print(type(self.training_config.optim_lr), self.training_config.optim_lr)
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.training_config.optim_lr
        )

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, RayDeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        else:
            distribute_kwargs = {}

        # Distributed model, optimizer, and scheduler
        self.model, self.optimizer, _ = self.strategy.distributed(
            self.model, self.optimizer, **distribute_kwargs
        )

    def custom_collate(self, batch):
        """Custom collate function to concatenate input tensors along their first dimension."""
        # Some batches contain None values,
        # if any files from the dataset did not match the criteria
        # (i.e. three auxilliary channels)
        batch = [x for x in batch if x is not None]

        return torch.cat(batch)

    def train(self, config, data):
        # Because of the way the ray cluster is set up, the strategy must be initialized within
        # the training function
        self.strategy.init()

        # Start the timer for profiling
        st = timer()

        self.training_config = VirgoTrainingConfiguration(**config)

        self.create_model_loss_optimizer()

        self.create_dataloaders(
            train_dataset=data[0],
            validation_dataset=data[1],
            test_dataset=data[2],
            collate_fn=self.custom_collate,
        )

        self.initialize_logger(hyperparams=config, rank=self.strategy.global_rank())

        if self.strategy.is_main_worker:
            print("TIMER: broadcast:", timer() - st, "s")
            print("\nDEBUG: start training")
            print("--------------------------------------------------------")

        loss_plot = []
        val_loss_plot = []
        acc_plot = []
        val_acc_plot = []
        best_val_loss = float("inf")

        for epoch in tqdm(range(self.training_config.num_epochs)):
            # lt = timer()

            if self.strategy.global_world_size() > 1:
                self.set_epoch(epoch)

            t_list = []
            st = time.time()
            epoch_loss = []

            for i, batch in enumerate(self.train_dataloader):
                t = timer()
                # The TensorDataset returns batches as lists of length 1
                if isinstance(batch, list):
                    batch = batch[0]

                if isinstance(self.strategy, RayDDPStrategy):
                    target = batch[:, 0].unsqueeze(1)
                    input = batch[:, 1:]
                else:
                    target = batch[:, 0].unsqueeze(1).to(self.device)
                    input = batch[:, 1:].to(self.device)

                target = target.float()

                self.optimizer.zero_grad()
                generated = self.model(input.float())
                loss = self.loss(generated, target)
                loss.backward()
                self.optimizer.step()
                epoch_loss.append(loss.detach().cpu().numpy())
                t_list.append(timer() - t)
                # itwinai - log loss as metric
                self.log(
                    loss.detach().cpu().numpy(),
                    "epoch_loss_batch",
                    kind="metric",
                    step=epoch * len(self.train_dataloader) + i,
                    batch_idx=i,
                )

            if self.strategy.is_main_worker:
                print("TIMER: train time", sum(t_list) / len(t_list), "s")
            val_loss = []

            for i, batch in enumerate(self.validation_dataloader):
                if isinstance(batch, list):
                    batch = batch[0]

                if isinstance(self.strategy, RayDDPStrategy):
                    target = batch[:, 0].unsqueeze(1)
                    input = batch[:, 1:]
                else:
                    target = batch[:, 0].unsqueeze(1).to(self.device)
                    input = batch[:, 1:].to(self.device)

                target = target.float()
                with torch.no_grad():
                    generated = self.model(input.float())
                    # generated=normalize_(generated,1)
                    loss = self.loss(generated, target)
                val_loss.append(loss.detach().cpu().numpy())
                # itwinai -log loss as metric
                self.log(
                    loss.detach().cpu().numpy(),
                    "val_loss_batch",
                    kind="metric",
                    step=epoch * len(self.validation_dataloader) + i,
                    batch_idx=i,
                )

            loss_plot.append(np.mean(epoch_loss))
            val_loss_plot.append(np.mean(val_loss))

            # itwinai - Log metrics/losses
            self.log(np.mean(epoch_loss), "epoch_loss", kind="metric", step=epoch)
            self.log(np.mean(val_loss), "val_loss", kind="metric", step=epoch)

            et = time.time()
            # itwinai - print() in a multi-worker context (distributed)
            if self.strategy.is_main_worker:
                print(
                    "epoch: {} loss: {} val loss: {} time:{}s".format(
                        epoch, loss_plot[-1], val_loss_plot[-1], et - st
                    )
                )

                # uncomment the following if you want to save checkpoint every
                # 100 epochs regardless of the performance of the model
                # checkpoint = {
                #     'epoch': epoch,
                #     'model_state_dict': generator.state_dict(),
                #     'optim_state_dict': optimizer.state_dict(),
                #     'loss': loss_plot[-1],
                #     'val_loss': val_loss_plot[-1],
                # }
                # if self.strategy.is_main_worker:
                #     # Save only in the main worker
                #     checkpoint_filename = checkpoint_path.format(epoch)
                #     torch.save(checkpoint, checkpoint_filename)

                # Average loss among all workers
                # itwinai - gather local loss from all the workers
            # worker_val_losses = self.strategy.gather_obj(val_loss_plot[-1])

            checkpoint = None
            if self.strategy.is_main_worker:
                # save checkpoint only if it is better than
                # the previous ones
                if self.training_config["save_best"] and val_loss_plot[-1] < best_val_loss:
                    # create checkpoint
                    checkpoint = {
                        "epoch": epoch,
                        "model_state_dict": self.model.state_dict(),
                        "optim_state_dict": self.optimizer.state_dict(),
                        "loss": loss_plot[-1],
                        "val_loss": val_loss_plot[-1],
                    }

            metrics = {"loss": val_loss_plot[-1]}
            self.checkpoint_and_report(
                epoch, tuning_metrics=metrics, checkpointing_data=checkpoint
            )

        self.close_logger()

        return loss_plot, val_loss_plot, acc_plot, val_acc_plot



### MAIN SETTINGS ###

generate_data = True

### DATA GENERATION SETTINGS ###

tag = "train_v0_"
num_payloads = 50
plot_a_example = False
num_cpus = 1
param_folder = "syn_param/"
payload_folder = "syn_payload/"
image_shape = (128, 128)

### PATH SETTINGS ###

image_directory_npy = "./test_data/joint_dataset_8_classes_real_remix_128x128.npy"
label_directory_npy = "./test_data/joint_dataset_8_classes_real_remix_labels.npy"
image_tag = "train_v0_*_payload_detected.json"  #: '*' in the name is the index place holder of a image in the image set
image_directory = payload_folder
mask_tag = "train_v0_*_payload_flux.json"
mask_directory = payload_folder

unet_dir = "./models/trained_UNet_test_v0.pt"
fcnn_dir = "./models/trained_Filter_test_v0.pt"
cnn1d_dir = "./models/trained_CNN1D_test_v0.pt"
offset = 0
size_of_set = 10

### SPLITTER SETTINGS ###

train_proportion = 0.8
validation_proportion = 0.1
test_proportion = 0.1
rnd_seed = 42
spl_name = "splitter"

### IMAGE/MASK ENGINE SETTINGS ###

engine_settings_unet = {
    "image": {
        "do_rot_phase_avg": True,
        "do_resize": True,
        "resize_size": image_shape,
        "do_binarize": False,
    },
    "mask": {
        "do_rot_phase_avg": True,
        "do_resize": True,
        "resize_size": image_shape,
        "do_binarize": True,
        "binarize_func": "thresh",
    },
}
engine_settings_filtercnn = {
    "image": {
        "do_rot_phase_avg": True,
        "do_resize": True,
        "resize_size": image_shape,
        "do_binarize": False,
    },
    "mask": {
        "do_rot_phase_avg": True,
        "do_resize": True,
        "resize_size": image_shape,
        "do_binarize": True,
        "binarize_func": "thresh",
    },
    "mask_maker": {
        "model": "UNet",
        "trained_image_to_mask_network_path": "./models/trained_UNet_test_v0.pt",
    },
}
engine_settings_cnn1d = {
    "mask": {
        "do_rot_phase_avg": True,
        "do_resize": True,
        "resize_size": image_shape,
        "do_binarize": True,
        "binarize_func": "thresh",
    }
}

### TRAINER CONFIGURATION ###

config = {
    "generator": "simple",
    "batch_size": 10,
    "optimizer": "adam",
    "optim_lr": 0.001,
    # "loss"                    : 'cross_entropy',  #"cross_entropy"
    "save_best": True,
    "shuffle_train": True,
    # "shuffle_validation"      : True,
    # "shuffle_test"            : True,
    "device": "auto",
    "num_workers_dataloader": 0,
    # "pin_gpu_memory"          : True,
}



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List

from pydantic import BaseModel

from itwinai.slurm.slurm_constants import JUWELS_HPC_MODULES, SLURM_TEMPLATE
from itwinai.slurm.utils import (
    get_slurm_job_parser,
    remove_indentation_from_multiline_string,
)


class SlurmScriptConfiguration(BaseModel):
    """Configuration object for the SLURM script. It contains all the settings for the
    SLURM script such as which hardware you are requesting or for how long to run it.
    As it allows for any ``pre_exec_command`` and ``exec_command``, it should work for
    any SLURM script.
    """

    # Settings for the SLURM configuration
    job_name: str | None = None
    account: str
    partition: str
    time: str

    std_out: Path | None = None
    err_out: Path | None = None

    num_nodes: int
    num_tasks_per_node: int
    gpus_per_node: int
    cpus_per_gpu: int

    # Tyipcally used to set up the environment before executing the command,
    # e.g. "ml Python", "source .venv/bin/activate" etc.
    pre_exec_command: str | None = None

    # Command to execute, typically an 'srun' command
    exec_command: str | None = None

    def format_script(self) -> str:
        """Uses the provided configuration parameters and formats a SLURM script with
        the requested settings.

        Returns:
            str: A string containing the SLURM script.
        """
        return SLURM_TEMPLATE.format_map(self.model_dump())


class SlurmScriptBuilder:
    """Builds a SLURM script for various itwinai use cases, typically machine-learning
    use cases. Uses the provided ``SlurmScriptConfiguration`` to build the script. Some
    methods, such as the run-all or scaling-test methods, will modify the given
    configuration in-place, so be careful if reusing the configuration object.
    """

    def __init__(
        self,
        slurm_script_configuration: SlurmScriptConfiguration,
        distributed_strategy: str,
        pre_exec_command: str | None = None,
        training_command: str | None = None,
        python_venv: str = ".venv",
        debug: bool = False,
        config_name: str = "config",
        config_path: str = ".",
        pipe_key: str = "training_pipeline",
        file_folder: Path = Path("slurm_scripts"),
        py_spy_profiling: bool = False,
        profiling_sampling_rate: int = 10,
    ):
        self.slurm_script_configuration = slurm_script_configuration
        self.distributed_strategy = distributed_strategy
        self.pre_exec_command = pre_exec_command
        self.training_command = training_command

        self.python_venv = python_venv
        self.debug = debug
        self.file_folder = file_folder

        self.py_spy_profiling = py_spy_profiling
        self.profiling_sampling_rate = profiling_sampling_rate

        # exec-pipeline-specific commands
        self.config_name = config_name
        self.config_path = config_path
        self.pipe_key = pipe_key

        if self.slurm_script_configuration.cpus_per_gpu > 0:
            self.omp_num_threads = self.slurm_script_configuration.cpus_per_gpu
        else:
            self.omp_num_threads = 1

    @property
    def training_cmd_formatter(self) -> Dict[str, str]:
        return {
            "dist_strat": self.distributed_strategy,
            "config_name": self.config_name,
            "config_path": self.config_path,
            "pipe_key": self.pipe_key,
        }

    def generate_identifier(self) -> str:
        num_nodes = self.slurm_script_configuration.num_nodes
        gpus_per_node = self.slurm_script_configuration.gpus_per_node
        return f"{self.distributed_strategy}-{num_nodes}x{gpus_per_node}"

    def get_training_command(self) -> str:
        if self.training_command:
            return self.training_command.format(**self.training_cmd_formatter)

        # Default for the TorchTrainer
        default_command = rf"""
            $(which itwinai) exec-pipeline \
            strategy={self.distributed_strategy} \
            checkpoints_location=checkpoints_{self.distributed_strategy}
        """
        default_command = default_command.strip()
        return remove_indentation_from_multiline_string(default_command)

    def get_pre_exec_command(self) -> str:
        """Generates a pre-execution command for the SLURM script. This will load the
        standard HPC modules, source the given python venv and set the OpenMP number
        of threads. Will also add debug echo statements if debug flag is set to True."""
        pre_exec_command = rf"""
            ml {" ".join(JUWELS_HPC_MODULES)}
            source {self.python_venv}/bin/activate
            export OMP_NUM_THREADS={self.omp_num_threads}
        """

        if self.pre_exec_command:
            pre_exec_command += f"\n{self.pre_exec_command}"
        if self.debug:
            pre_exec_command += "\n" + self.get_debug_command()

        pre_exec_command = pre_exec_command.strip()
        return remove_indentation_from_multiline_string(pre_exec_command)

    def get_srun_command(self, torch_log_dir: str = "logs_torchrun") -> str:
        """Generates an srun command for the different distributed ML frameworks using
        the internal training command. Sets up rendezvous connections etc. when
        necessary and exports necessary environment variables.
        """

        if self.distributed_strategy in ["ddp", "deepspeed"]:
            rdzv_endpoint = (
                "'$(scontrol show hostnames \"$SLURM_JOB_NODELIST\" | head -n 1)'i:29500"
            )
            bash_command = rf"""torchrun \
                --log_dir='{torch_log_dir}' \
                --nnodes=$SLURM_NNODES \
                --nproc_per_node=$SLURM_GPUS_PER_NODE \
                --rdzv_id=$SLURM_JOB_ID \
                --rdzv_conf=is_host=\$(((SLURM_NODEID)) && echo 0 || echo 1) \
                --rdzv_backend=c10d \
                --rdzv_endpoint={rdzv_endpoint} \
                {self.get_training_command()}"""
            if self.py_spy_profiling:
                # Prepending the py-spy profiling command
                py_spy_profiling_output = "py_spy_profiling_$SLURM_NODEID.txt"
                bash_command = (
                    f"py-spy record -r {self.profiling_sampling_rate} -s "
                    f"-o {py_spy_profiling_output} -f raw -- " + bash_command
                )

            main_command = rf"""
            srun --cpu-bind=none --ntasks-per-node=1 \
            bash -c "{bash_command}"
        """
        else:
            gpus_per_node = self.slurm_script_configuration.gpus_per_node
            num_nodes = self.slurm_script_configuration.num_nodes
            num_tasks = gpus_per_node * num_nodes

            # E.g. "0,1,2,3" with 4 GPUs per node
            cuda_visible_devices = ",".join(str(i) for i in range(gpus_per_node))

            main_command = rf"""
            export CUDA_VISIBLE_DEVICES="{cuda_visible_devices}"
            srun --cpu-bind=none \
                --ntasks-per-node=$SLURM_GPUS_PER_NODE \
                --cpus-per-task=$SLURM_CPUS_PER_GPU \
                --ntasks={num_tasks} \
                {self.get_training_command()}
        """

        main_command = main_command.strip()
        return remove_indentation_from_multiline_string(main_command)

    def get_debug_command(self) -> str:
        """Generates a debug command to be added to the pre-exec command."""

        debug_print_command = rf"""
        echo "### SLURM Variables ###"
        echo "SLURM_SUBMIT_DIR: $SLURM_SUBMIT_DIR"
        echo "SLURM_JOB_ID: $SLURM_JOB_ID"
        echo "SLURM_JOB_NODELIST: $SLURM_JOB_NODELIST"
        echo "SLURM_NNODES: $SLURM_NNODES"
        echo "SLURM_NTASKS: $SLURM_NTASKS"
        echo "SLURM_TASKS_PER_NODE: $SLURM_TASKS_PER_NODE"
        echo "SLURM_SUBMIT_HOST: $SLURM_SUBMIT_HOST"
        echo "SLURMD_NODENAME: $SLURMD_NODENAME"
        echo "CUDA_VISIBLE_DEVICES: $CUDA_VISIBLE_DEVICES"
        echo ""
        echo "### Other Variables ###"
        echo "Distributed Strategy: {self.distributed_strategy}"
        echo "Current working directory: $(pwd)"
        echo "Which python: $(which python)"
        """
        debug_print_command = debug_print_command.strip()
        return remove_indentation_from_multiline_string(debug_print_command)

    def process_slurm_script(
        self,
        file_path: Path | None = None,
        save_script: bool = True,
        submit_slurm_job: bool = True,
    ) -> None:
        """Will generate and process a SLURM script according to specifications. Will
        run the script if ``submit_slurm_job`` is set to True, will store the SLURM script file
        if ``save_script`` is set to True. If both are false, it will simply print the
        script to the console without any further processing. Also creates the
        needed directories for the stderr and stdout for the script.

        Args:
            file_path: Where to store the file before processing the script. Also the location
                it will remain if ``save_script`` is set to True.
            save_script: Whether to keep or delete the file after finishing processing.
            submit_slurm_job: Whether to submit the script as a SLURM job.
        """
        job_identifier = self.generate_identifier()
        self.slurm_script_configuration.pre_exec_command = self.get_pre_exec_command()
        self.slurm_script_configuration.exec_command = self.get_srun_command()

        # Setting some default fields
        if self.slurm_script_configuration.job_name is None:
            self.slurm_script_configuration.job_name = job_identifier

        if self.slurm_script_configuration.std_out is None:
            std_out_path = Path("slurm_job_logs") / f"{job_identifier}.out"
            self.slurm_script_configuration.std_out = std_out_path
        if self.slurm_script_configuration.err_out is None:
            err_out_path = Path("slurm_job_logs") / f"{job_identifier}.err"
            self.slurm_script_configuration.err_out = err_out_path

        # Making sure the std out and err out folders exist
        self.slurm_script_configuration.std_out.parent.mkdir(exist_ok=True, parents=True)
        self.slurm_script_configuration.err_out.parent.mkdir(exist_ok=True, parents=True)

        # Generate the script using the given configuration
        script = self.slurm_script_configuration.format_script()
        if not submit_slurm_job and not save_script:
            upper_banner_str = f"{'#' * 20} SLURM Script Preview {'#' * 20}"
            print(upper_banner_str)
            print(script)
            print("#" * len(upper_banner_str))
            return

        temp_dir = None
        if save_script:
            file_path = file_path or self.file_folder / f"{job_identifier}.sh"
            if file_path.exists():
                raise FileExistsError(
                    f"File '{file_path.resolve()}' already exists! Give a different path "
                    f"or delete the file first."
                )

            self.file_folder.mkdir(exist_ok=True, parents=True)
            print(f"Storing SLURM script at '{file_path.resolve()}'.")
        else:
            temp_dir = TemporaryDirectory()
            file_path = Path(temp_dir.name) / f"{job_identifier}.sh"

        with open(file_path, "w") as f:
            f.write(script)

        if submit_slurm_job:
            subprocess.run(["sbatch", str(file_path.resolve())])

        if temp_dir:
            temp_dir.cleanup()

    def run_slurm_script_all_strategies(
        self,
        file_folder: Path = Path("slurm_scripts"),
        save_script: bool = True,
        submit_slurm_job: bool = True,
        strategies: List[str] = ["ddp", "horovod", "deepspeed"],
    ):
        """Runs the SLURM script with all the given strategies. Does the same as the
        ``runall.sh`` script has been doing.
        """
        self.file_folder = file_folder
        for strategy in strategies:
            self.distributed_strategy = strategy
            job_identifier = self.generate_identifier()

            # Overriding job_name, std_out and err_out
            self.slurm_script_configuration.job_name = job_identifier
            std_out_path = Path("slurm_job_logs") / f"{job_identifier}.out"
            err_out_path = Path("slurm_job_logs") / f"{job_identifier}.err"
            self.slurm_script_configuration.std_out = std_out_path
            self.slurm_script_configuration.err_out = err_out_path

            self.process_slurm_script(
                save_script=save_script, submit_slurm_job=submit_slurm_job
            )

    def run_scaling_test(
        self,
        file_folder: Path = Path("slurm_scripts"),
        save_script: bool = True,
        submit_slurm_job: bool = True,
        strategies: List[str] = ["ddp", "horovod", "deepspeed"],
        num_nodes_list: List[int] = [1, 2, 4, 8],
    ):
        """Runs a scaling test, i.e. runs all the strategies with separate runs for each
        distinct number of nodes.
        """
        for num_nodes in num_nodes_list:
            self.slurm_script_configuration.num_nodes = num_nodes
            self.run_slurm_script_all_strategies(
                file_folder=file_folder,
                save_script=save_script,
                submit_slurm_job=submit_slurm_job,
                strategies=strategies,
            )


def generate_default_slurm_script() -> None:
    """Generates and optionally submits a default SLURM script.

    This function creates a SLURM script using the `SlurmScriptBuilder`, based on
    command-line arguments parsed from `get_slurm_job_parser()`. It sets up a
    basic SLURM configuration with common parameters like job name, account,
    requested resources, and execution commands.

    If `--no-submit-job` is provided, the script will not be submitted via `sbatch`.
    If `--no-save-script` is provided, the generated SLURM script will be deleted
    after execution.
    """
    parser = get_slurm_job_parser()
    args = parser.parse_args()

    slurm_script_configuration = SlurmScriptConfiguration(
        job_name=args.job_name,
        account=args.account,
        time=args.time,
        partition=args.partition,
        std_out=args.std_out,
        err_out=args.err_out,
        num_nodes=args.num_nodes,
        num_tasks_per_node=args.num_tasks_per_node,
        gpus_per_node=args.gpus_per_node,
        cpus_per_gpu=args.cpus_per_gpu,
    )

    slurm_script_builder = SlurmScriptBuilder(
        slurm_script_configuration=slurm_script_configuration,
        distributed_strategy=args.dist_strat,
        debug=args.debug,
        python_venv=args.python_venv,
        pre_exec_command=args.pre_exec_cmd,
        training_command=args.training_cmd,
        config_path=args.config_path,
        config_name=args.config_name,
        pipe_key=args.pipe_key,
        py_spy_profiling=args.py_spy,
        profiling_sampling_rate=args.profiling_rate,
    )

    submit_job = not args.no_submit_job
    save_script = not args.no_save_script

    mode = args.mode
    if mode == "single":
        slurm_script_builder.process_slurm_script(
            submit_slurm_job=submit_job, save_script=save_script
        )
    elif mode == "runall":
        slurm_script_builder.run_slurm_script_all_strategies(
            submit_slurm_job=submit_job, save_script=save_script
        )
    elif mode == "scaling-test":
        slurm_script_builder.run_scaling_test(
            submit_slurm_job=submit_job,
            save_script=save_script,
            num_nodes_list=args.scalability_nodes,
        )



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from pathlib import Path
from typing import Set

import pandas as pd

from itwinai.scalability_report.utils import check_contains_columns


def read_scalability_metrics_from_csv(
    data_dir: Path | str, expected_columns: Set
) -> pd.DataFrame:
    """Reads and validates scalability metric CSV files from a directory and combines
    them into a single DataFrame.

    Args:
        data_dir (Path | str): Path to the directory containing the CSV files. All files
            in the directory must have a .csv extension.
        expected_columns (Set): A set of column names expected to be present in each CSV
            file.

    Returns:
        pd.DataFrame: A DataFrame containing the concatenated data from all valid CSV
        files in the directory.

    Raises:
        ValueError: If the directory contains non-CSV files, if no .csv files are found,
        or if any file is missing the expected columns.
    """
    if isinstance(data_dir, str):
        data_dir = Path(data_dir)

    file_paths = list(data_dir.iterdir())

    # Checking that all files end with .csv
    if len([f for f in file_paths if f.suffix != ".csv"]) > 0:
        raise ValueError(
            f"Directory '{data_dir.resolve()} contains files with suffix different "
            f"from .csv!"
        )

    if len(file_paths) == 0:
        raise ValueError(f"Found no .csv files in directory: '{data_dir.resolve()}'")

    dataframes = []
    for file_path in file_paths:
        df = pd.read_csv(file_path)
        check_contains_columns(df=df, expected_columns=expected_columns, file_path=file_path)
        dataframes.append(df)

    return pd.concat(dataframes)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Unknown
# --------------------------------------------------------------------------------------

import argparse
import sys
from pathlib import Path
from time import time

import h5py
import numpy as np
import pandas as pd
from gwpy.timeseries import TimeSeries
from tqdm import tqdm


def append_to_hdf5_dataset(
    file_path: Path,
    dataset_name: str,
    array: np.ndarray,
    expected_datapoint_shape: tuple,
):
    if tuple(array.shape[1:]) != expected_datapoint_shape:
        actual_shape_str = ", ".join(str(s) for s in array.shape)
        expected_shape_str = ", ".join(str(s) for s in expected_datapoint_shape)
        raise ValueError(
            f"'array' has an incorrect shape: ({actual_shape_str}). "
            f"Should have been (x, {expected_shape_str})."
        )

    print(f"Appending to file: '{str(file_path.resolve())}'.")
    with h5py.File(file_path, "a") as f:
        dset=f[dataset_name]
        dset.resize(dset.shape[0] + array.shape[0], axis=0)
        dset[-array.shape[0] :] = array


def generate_hdf5_dataset(
    output_file: str = "virgo_data.hdf5",
    dataset_name: str = "virgo_dataset",
    num_datapoints=5,
    duration=10,
    sample_rate=500,
    num_aux_channels=10,
    num_waves_range=(10, 15),
    noise_amplitude=0.1,
    num_processes=4,
    square_size=64,
    datapoints_per_file=10,
    seed=None,
) -> None:
    """Generate a folder with num_files pickle files containing synthetic gravitational
    waves data.

    Args:
        folder_name (string): the path and name where the files will be stored
        num_files (int): Number of files which will be created.
        duration (float): Duration of the time series data in seconds (default is 6
            seconds).
        num_aux_channels (int): Number of auxiliary channels, containing the data from
            the auxiliary sensors in the detector which do not go into the strain.
        sample_rate (int): Sampling rate of the time series data in samples per second
            (default is 500 samples per second).
        num_waves_range (tuple): Range for the random number of sine waves to be
            generated for each time series. Format: (min_num_waves, max_num_waves)
            (default is (10, 15)).
        noise_amplitude (float): Amplitude of the smooth random noise added to the
            time series data (default is 0.1).
        num_processes (int): Number of cores for multiprocess (default 20)
        square_size (int): Size in pixels of qplot image (default is 500 samples per
            second).
        datapoints_per_file (int): number of independent datapoints per pickle file.
    """
    sys.path.append(str(Path.cwd().resolve()))
    from src.dataset import generate_cut_image_dataset

    if seed is None:
        # Since we can't retrieve the seed from numpy, we set it so that we can
        # store it in the dataset. More information:
        # https://stackoverflow.com/questions/32172054/how-can-i-retrieve-the-current-seed-of-numpys-random-number-generator
        seed = np.random.randint(0, 2**32, dtype=np.uint64)
    np.random.seed(seed)

    datapoints = []

    # Creating empty HDF5 file
    datapoint_shape = (num_aux_channels + 1, square_size, square_size)
    output_file_path = Path(output_file)
    output_file_path.parent.mkdir(parents=True, exist_ok=True)
    print(f"Creating/overwriting file: '{output_file_path.resolve()}'.")
    with h5py.File(output_file_path, "w") as f:
        dataset = f.create_dataset(
            dataset_name,
            shape=(0, *datapoint_shape),
            maxshape=(None, *datapoint_shape),
            dtype=np.float32,
        )
        dataset.attrs["Description"] = (
            "Synthetic time series data for the Virgo use case"
        )
        dataset.attrs["main_channel_idx"] = 0
        dataset.attrs["seed"] = str(seed)

    for f in tqdm(range(num_datapoints)):
        times = np.linspace(0, duration, duration * sample_rate)
        # Initialise the main data as a list of zeros
        main_data = np.zeros(len(times))
        dictionary_aux = {}
        for i in range(num_aux_channels):
            channel_name = "Aux_" + str(i)

            # Initialize an array to store the generated wave data for this row
            wave_data = np.zeros(len(times))
            # Determine the number of sine waves to generate for this column randomly
            num_waves = np.random.randint(*num_waves_range)

            # Generate each sine wave
            for _ in range(num_waves):
                # Randomly generate parameters for the sine wave (amplitude, frequency, phase)
                amplitude = np.random.uniform(0.5, 2.0)
                frequency = np.random.uniform(0.5, 5.0)
                phase = np.random.uniform(0, 2 * np.pi)

                # Generate the sine wave and add it to the wave_data
                wave_data += amplitude * np.sin(2 * np.pi * frequency * times + phase)

            # Add smooth random noise to the wave data
            smooth_noise = np.random.normal(0, noise_amplitude, len(times))
            wave_data += smooth_noise

            coeff = np.random.rand()

            main_data += coeff * wave_data

            # Create a TimeSeries object from the wave data
            ts = TimeSeries(wave_data, t0=0, dt=1 / sample_rate)

            dictionary_aux[channel_name] = [ts]

        # Creating the main timeseries
        main_data += np.random.normal(0, noise_amplitude)
        ts_main = TimeSeries(main_data, dt=1 / sample_rate)

        main_entry = {"Main": [ts_main]}

        dictionary = {**main_entry, **dictionary_aux}

        # turn dictionary into dataframe
        df_ts = pd.DataFrame(dictionary)

        datapoints.append(df_ts)
        if len(datapoints) % datapoints_per_file == 0:
            # Converting to numpy array
            df_concat = pd.concat(datapoints)
            df_concat = generate_cut_image_dataset(
                df=df_concat,
                channels=list(df_ts.columns),
                num_processes=num_processes,
                square_size=square_size,
            )
            value_array = df_concat.to_numpy()
            value_array = np.stack(
                [np.stack(row) for row in value_array], dtype=np.float32
            )

            append_to_hdf5_dataset(
                file_path=output_file_path,
                dataset_name=dataset_name,
                array=value_array,
                expected_datapoint_shape=datapoint_shape,
            )
            datapoints = []

    # Adding any remaining datapoints
    if len(datapoints) > 0:
        df_concat = pd.concat(datapoints)
        df_concat = generate_cut_image_dataset(
            df=df_concat,
            channels=list(df_concat.columns),
            num_processes=num_processes,
            square_size=square_size,
        )
        value_array = df_concat.to_numpy()
        value_array = np.stack([np.stack(row) for row in value_array], dtype=np.float32)
        append_to_hdf5_dataset(
            file_path=output_file_path,
            dataset_name=dataset_name,
            array=value_array,
            expected_datapoint_shape=datapoint_shape,
        )
        datapoints = []


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Virgo Dataset Generation")
    parser.add_argument(
        "--num-datapoints", type=int, help="Number of datapoints to be created."
    )
    parser.add_argument(
        "--save-frequency",
        type=int,
        help="How often to save to file while generating. Also saves when finished. ",
        default=100,
    )
    parser.add_argument(
        "--num-processes",
        type=int,
        help="How many processes to use when generating data",
        default=1,
    )
    parser.add_argument(
        "--save-location",
        type=str,
        help="Location to save the resulting HDF5 file.",
        default="virgo_data.hdf5",
    )

    parser.add_argument("--seed", type=int, help="Seed for random number generator")

    args = parser.parse_args()

    start = time()
    generate_hdf5_dataset(
        output_file=args.save_location,
        num_datapoints=args.num_datapoints,
        duration=16,
        sample_rate=500,
        num_aux_channels=3,
        num_waves_range=(10, 15),
        noise_amplitude=0.5,
        datapoints_per_file=args.save_frequency,
        num_processes=args.num_processes,
        seed=args.seed,
    )
    end = time()
    total_time = end - start
    print(f"Generation took {total_time:.2f} seconds!")



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from unittest.mock import MagicMock, call, patch

import pytest

from itwinai.components import BaseComponent, monitor_exec
from itwinai.distributed import suppress_workers_print


class SilentComponent(BaseComponent):
    """Fake component class to test the decorator meant to suppress print()."""

    def __init__(self, max_items: int, name: str = "SilentComponent") -> None:
        super().__init__(name)
        self.save_parameters(max_items=max_items, name=name)
        self.max_items = max_items

    @suppress_workers_print
    def execute(self, foo, bar=123):
        print("Executing SilentComponent")
        return "silent component result"


class FakeComponent(BaseComponent):
    """Fake component class to test the use of two decorators at the same time."""

    def __init__(self, max_items: int, name: str = "FakeComponent") -> None:
        super().__init__(name)
        self.save_parameters(max_items=max_items, name=name)
        self.max_items = max_items

    @suppress_workers_print
    @monitor_exec
    def execute(self, foo, bar):
        print("Executing FakeComponent")
        return "fake component result"


@pytest.fixture
def mock_component():
    """Fixture to provide a mock BaseComponent."""
    component = MagicMock(spec=BaseComponent)
    component.name = "TestComponent"
    return component


@pytest.fixture
def mock_fake_component():
    """Fixture to provide a FakeComponent instance."""
    component = FakeComponent(max_items=10)
    component.cleanup = MagicMock(spec=FakeComponent.cleanup)
    return component


def test_suppress_workers_print_decorator():
    """Test suppress_workers_print decorator behavior."""
    with (
        patch("builtins.print", autospec=True) as mock_print,
        patch("itwinai.distributed.detect_distributed_environment", autospec=True) as mock_env,
        patch(
            "itwinai.distributed.distributed_patch_print", autospec=True
        ) as mock_patch_print,
    ):
        # Mock environment: global rank different from 0 (non-main worker)
        mock_env.return_value.global_rank = 1
        mock_patch_print.return_value = lambda *args, **kwargs: None  # Suppress print

        @suppress_workers_print
        def dummy_function():
            print("This should be suppressed")

        dummy_function()
        mock_print.assert_not_called()  # Ensure print was suppressed

        # Check that outside the decorated function print can still be used
        print("This should be printed")
        mock_print.assert_called_once_with("This should be printed")
        mock_print.reset_mock()

        # Mock environment: global rank is 0 (main worker)
        mock_env.return_value.global_rank = 0
        mock_patch_print.return_value = print  # Use default print

        dummy_function()
        mock_print.assert_called_once_with("This should be suppressed")


def test_suppress_workers_print_component():
    """Test suppress_workers_print decorator behavioron SilentComponent's
    execute method."""
    with (
        patch("builtins.print", autospec=True) as mock_print,
        patch("itwinai.distributed.detect_distributed_environment", autospec=True) as mock_env,
        patch(
            "itwinai.distributed.distributed_patch_print", autospec=True
        ) as mock_patch_print,
    ):
        # Initialize the SilentComponent instance
        silent_component = SilentComponent(max_items=10)

        # Mock environment: global rank different from 0 (non-main worker)
        mock_env.return_value.global_rank = 1
        mock_patch_print.return_value = lambda *args, **kwargs: None  # Suppress print

        # Call the execute method on a non-main worker
        result = silent_component.execute("foo", bar=123)

        # Ensure the execute method's print statement was suppressed
        mock_print.assert_not_called()
        assert result == "silent component result"

        # Check that outside the decorated function print can still be used
        print("This should be printed")
        mock_print.assert_called_once_with("This should be printed")
        mock_print.reset_mock()

        # Mock environment: global rank is 0 (main worker)
        mock_env.return_value.global_rank = 0
        mock_patch_print.return_value = print  # Use default print

        # Call the execute method on the main worker
        result = silent_component.execute("foo", bar=123)

        # Ensure the print statement in the execute method was executed correctly
        mock_print.assert_called_once_with("Executing SilentComponent")
        assert result == "silent component result"


def test_monitor_exec_decorator(mock_component):
    """Test monitor_exec decorator behavior."""
    with patch("time.time") as mock_time:
        mock_time.side_effect = [100.0, 105.0]  # Simulate 5 seconds execution time

        @monitor_exec
        def dummy_method(self):
            return "Execution result"

        result = dummy_method(mock_component)

        # Assert the result of the decorated function
        assert result == "Execution result"

        # Check that the start and end messages were logged
        mock_component._printout.assert_any_call("Starting execution of 'TestComponent'...")
        mock_component._printout.assert_any_call("'TestComponent' executed in 5.000s")

        # Ensure the cleanup method was called
        mock_component.cleanup.assert_called_once()

        # Check that execution time was set correctly
        assert mock_component.exec_t == 5.0


def test_combined_decorators_on_fake_component(mock_fake_component):
    """Test the combination of suppress_workers_print and monitor_exec decorators
    on FakeComponent."""

    with (
        patch("builtins.print", autospec=True) as mock_print,
        patch("itwinai.distributed.detect_distributed_environment", autospec=True) as mock_env,
        patch(
            "itwinai.distributed.distributed_patch_print", autospec=True
        ) as mock_patch_print,
        patch("time.time", autospec=True) as mock_time,
    ):
        # Simulate time progression for execution timing
        mock_time.side_effect = [100.0, 105.0]  # Simulate a 5-second execution time

        # Case 1: Non-main worker (print suppressed)
        mock_env.return_value.global_rank = 1
        mock_patch_print.return_value = lambda *args, **kwargs: None  # Suppress print

        result = mock_fake_component.execute("foo", "bar")
        assert result == "fake component result"

        # Ensure the print statement was suppressed for non-main worker
        mock_print.assert_not_called()

        assert mock_fake_component.exec_t == 5.0
        mock_fake_component.cleanup.assert_called_once()

        # Reset mock calls for next test case
        mock_print.reset_mock()
        mock_fake_component.cleanup.reset_mock()
        mock_time.side_effect = [100.0, 105.0]  # Simulate a 5-second execution time

        # Case 2: Main worker (print allowed)
        mock_env.return_value.global_rank = 0
        mock_patch_print.return_value = print  # Use standard print

        result = mock_fake_component.execute("foo", "bar")
        assert result == "fake component result"

        # Ensure the print statement was not suppressed for the main worker
        mock_print.assert_has_calls(
            [
                call("############################################"),
                call("# Starting execution of 'FakeComponent'... #"),
                call("############################################"),
                call("Executing FakeComponent"),
                call("######################################"),
                call("# 'FakeComponent' executed in 5.000s #"),
                call("######################################"),
            ]
        )

        assert mock_fake_component.exec_t == 5.0
        mock_fake_component.cleanup.assert_called_once()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Kalliopi Tsolaki
#
# Credit:
# - Kalliopi Tsolaki <kalliopi.tsolaki@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


import math
import os
import sys
from collections import defaultdict
from typing import Any

import lightning as pl
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from itwinai.loggers import Logger as BaseItwinaiLogger


class Generator(nn.Module):
    def __init__(self, latent_dim):  # img_shape
        super().__init__()
        # self.img_shape = img_shape
        self.latent_dim = latent_dim

        self.l1 = nn.Linear(self.latent_dim, 5184)
        self.up1 = nn.Upsample(scale_factor=(6, 6, 6), mode="trilinear", align_corners=False)
        self.conv1 = nn.Conv3d(in_channels=8, out_channels=8, kernel_size=(6, 6, 8), padding=0)
        nn.init.kaiming_uniform_(self.conv1.weight)
        # num_features is the number of channels (see doc)
        self.bn1 = nn.BatchNorm3d(num_features=8, eps=1e-6)
        self.pad1 = nn.ConstantPad3d((1, 1, 2, 2, 2, 2), 0)

        self.conv2 = nn.Conv3d(in_channels=8, out_channels=6, kernel_size=(4, 4, 6), padding=0)
        nn.init.kaiming_uniform_(self.conv2.weight)
        self.bn2 = nn.BatchNorm3d(num_features=6, eps=1e-6)
        self.pad2 = nn.ConstantPad3d((1, 1, 2, 2, 2, 2), 0)

        self.conv3 = nn.Conv3d(in_channels=6, out_channels=6, kernel_size=(4, 4, 6), padding=0)
        nn.init.kaiming_uniform_(self.conv3.weight)
        self.bn3 = nn.BatchNorm3d(num_features=6, eps=1e-6)
        self.pad3 = nn.ConstantPad3d((1, 1, 2, 2, 2, 2), 0)

        self.conv4 = nn.Conv3d(in_channels=6, out_channels=6, kernel_size=(4, 4, 6), padding=0)
        nn.init.kaiming_uniform_(self.conv4.weight)
        self.bn4 = nn.BatchNorm3d(num_features=6, eps=1e-6)
        self.pad4 = nn.ConstantPad3d((0, 0, 1, 1, 1, 1), 0)

        self.conv5 = nn.Conv3d(in_channels=6, out_channels=6, kernel_size=(3, 3, 5), padding=0)
        nn.init.kaiming_uniform_(self.conv5.weight)
        self.bn5 = nn.BatchNorm3d(num_features=6, eps=1e-6)
        self.pad5 = nn.ConstantPad3d((0, 0, 1, 1, 1, 1), 0)

        self.conv6 = nn.Conv3d(in_channels=6, out_channels=6, kernel_size=(3, 3, 3), padding=0)
        nn.init.kaiming_uniform_(self.conv6.weight)

        self.conv7 = nn.Conv3d(in_channels=6, out_channels=1, kernel_size=(2, 2, 2), padding=0)
        nn.init.xavier_normal_(self.conv7.weight)

    def forward(self, z):
        img = self.l1(z)
        img = img.view(-1, 8, 9, 9, 8)
        img = self.up1(img)
        img = self.conv1(img)
        img = F.relu(img)
        img = self.bn1(img)
        img = self.pad1(img)

        img = self.conv2(img)
        img = F.relu(img)
        img = self.bn2(img)
        img = self.pad2(img)

        img = self.conv3(img)
        img = F.relu(img)
        img = self.bn3(img)
        img = self.pad3(img)

        img = self.conv4(img)
        img = F.relu(img)
        img = self.bn4(img)
        img = self.pad4(img)

        img = self.conv5(img)
        img = F.relu(img)
        img = self.bn5(img)
        img = self.pad5(img)

        img = self.conv6(img)
        img = F.relu(img)

        img = self.conv7(img)
        img = F.relu(img)

        return img


class Discriminator(nn.Module):
    def __init__(self, power):
        super().__init__()

        self.power = power

        self.conv1 = nn.Conv3d(
            in_channels=1, out_channels=16, kernel_size=(5, 6, 6), padding=(2, 3, 3)
        )
        self.drop1 = nn.Dropout(0.2)
        self.pad1 = nn.ConstantPad3d((1, 1, 0, 0, 0, 0), 0)

        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=8, kernel_size=(5, 6, 6), padding=0
        )
        self.bn1 = nn.BatchNorm3d(num_features=8, eps=1e-6)
        self.drop2 = nn.Dropout(0.2)
        self.pad2 = nn.ConstantPad3d((1, 1, 0, 0, 0, 0), 0)

        self.conv3 = nn.Conv3d(in_channels=8, out_channels=8, kernel_size=(5, 6, 6), padding=0)
        self.bn2 = nn.BatchNorm3d(num_features=8, eps=1e-6)
        self.drop3 = nn.Dropout(0.2)

        self.conv4 = nn.Conv3d(in_channels=8, out_channels=8, kernel_size=(5, 6, 6), padding=0)
        self.bn3 = nn.BatchNorm3d(num_features=8, eps=1e-6)
        self.drop4 = nn.Dropout(0.2)

        self.avgpool = nn.AvgPool3d((2, 2, 2))
        self.flatten = nn.Flatten()

        # The input features for the Linear layer need to be calculated based
        # on the output shape from the previous layers.
        self.fakeout = nn.Linear(19152, 1)
        self.auxout = nn.Linear(19152, 1)  # The same as above for this layer.

    # calculate sum of intensities
    def ecal_sum(self, image, daxis):
        sum = torch.sum(image, dim=daxis)
        return sum

    # angle calculation
    def ecal_angle(self, image, daxis1):
        image = torch.squeeze(image, dim=daxis1)  # squeeze along channel axis

        # get shapes
        x_shape = image.shape[1]
        y_shape = image.shape[2]
        z_shape = image.shape[3]
        sumtot = torch.sum(image, dim=(1, 2, 3))  # sum of events

        # get 1. where event sum is 0 and 0 elsewhere
        amask = torch.where(sumtot == 0.0, torch.ones_like(sumtot), torch.zeros_like(sumtot))
        # masked_events = torch.sum(amask)  # counting zero sum events

        # ref denotes barycenter as that is our reference point
        x_ref = torch.sum(
            torch.sum(image, dim=(2, 3))
            * (
                torch.arange(x_shape, device=image.device, dtype=torch.float32).unsqueeze(0)
                + 0.5
            ),
            dim=1,
        )  # sum for x position * x index
        y_ref = torch.sum(
            torch.sum(image, dim=(1, 3))
            * (
                torch.arange(y_shape, device=image.device, dtype=torch.float32).unsqueeze(0)
                + 0.5
            ),
            dim=1,
        )
        z_ref = torch.sum(
            torch.sum(image, dim=(1, 2))
            * (
                torch.arange(z_shape, device=image.device, dtype=torch.float32).unsqueeze(0)
                + 0.5
            ),
            dim=1,
        )

        # return max position if sumtot=0 and divide by sumtot otherwise
        x_ref = torch.where(sumtot == 0.0, torch.ones_like(x_ref), x_ref / sumtot)
        y_ref = torch.where(sumtot == 0.0, torch.ones_like(y_ref), y_ref / sumtot)
        z_ref = torch.where(sumtot == 0.0, torch.ones_like(z_ref), z_ref / sumtot)

        # reshape
        x_ref = x_ref.unsqueeze(1)
        y_ref = y_ref.unsqueeze(1)
        z_ref = z_ref.unsqueeze(1)

        sumz = torch.sum(image, dim=(1, 2))  # sum for x,y planes going along z

        # Get 0 where sum along z is 0 and 1 elsewhere
        zmask = torch.where(sumz == 0.0, torch.zeros_like(sumz), torch.ones_like(sumz))

        x = torch.arange(x_shape, device=image.device).unsqueeze(0)  # x indexes
        x = (x.unsqueeze(2).float()) + 0.5
        y = torch.arange(y_shape, device=image.device).unsqueeze(0)  # y indexes
        y = (y.unsqueeze(2).float()) + 0.5

        # barycenter for each z position
        x_mid = torch.sum(torch.sum(image, dim=2) * x, dim=1)
        y_mid = torch.sum(torch.sum(image, dim=1) * y, dim=1)

        x_mid = torch.where(
            sumz == 0.0, torch.zeros_like(sumz), x_mid / sumz
        )  # if sum != 0 then divide by sum
        y_mid = torch.where(
            sumz == 0.0, torch.zeros_like(sumz), y_mid / sumz
        )  # if sum != 0 then divide by sum

        # Angle Calculations
        z = (
            torch.arange(
                z_shape,
                device=image.device,
                dtype=torch.float32,
                # Make an array of z indexes for all events
            )
            + 0.5
        ) * torch.ones_like(z_ref)

        # projection from z axis with stability check
        zproj = torch.sqrt(
            torch.max(
                (x_mid - x_ref) ** 2.0 + (z - z_ref) ** 2.0,
                torch.tensor([torch.finfo(torch.float32).eps]).to(x_mid.device),
            )
        )
        # torch.finfo(torch.float32).eps))
        # to avoid divide by zero for zproj =0
        m = torch.where(zproj == 0.0, torch.zeros_like(zproj), (y_mid - y_ref) / zproj)
        m = torch.where(z < z_ref, -1 * m, m)  # sign inversion
        ang = (math.pi / 2.0) - torch.atan(m)  # angle correction
        zmask = torch.where(zproj == 0.0, torch.zeros_like(zproj), zmask)
        ang = ang * zmask  # place zero where zsum is zero
        ang = ang * z  # weighted by position
        sumz_tot = z * zmask  # removing indexes with 0 energies or angles

        # zunmasked = K.sum(zmask, axis=1) # used for simple mean
        # Mean does not include positions where zsum=0
        # ang = K.sum(ang, axis=1)/zunmasked

        # sum ( measured * weights)/sum(weights)
        ang = torch.sum(ang, dim=1) / torch.sum(sumz_tot, dim=1)
        # Place 100 for measured angle where no energy is deposited in events
        ang = torch.where(amask == 0.0, ang, 100.0 * torch.ones_like(ang))
        ang = ang.unsqueeze(1)
        return ang

    def forward(self, x):
        z = self.conv1(x)
        z = F.leaky_relu(z)
        z = self.drop1(z)
        z = self.pad1(z)

        z = self.conv2(z)
        z = F.leaky_relu(z)
        z = self.bn1(z)
        z = self.drop2(z)
        z = self.pad2(z)

        z = self.conv3(z)
        z = F.leaky_relu(z)
        z = self.bn2(z)
        z = self.drop3(z)

        z = self.conv4(z)
        z = F.leaky_relu(z)
        z = self.bn3(z)
        z = self.drop4(z)
        z = self.avgpool(z)
        z = self.flatten(z)

        # generation output that says fake/real
        fake = torch.sigmoid(self.fakeout(z))
        aux = self.auxout(z)  # auxiliary output
        inv_image = x.pow(1.0 / self.power)
        ang = self.ecal_angle(inv_image, 1)  # angle calculation
        ecal = self.ecal_sum(inv_image, (2, 3, 4))  # sum of energies

        return fake, aux, ang, ecal


class ThreeDGAN(pl.LightningModule):
    def __init__(
        self,
        latent_size=256,
        loss_weights=[3, 0.1, 25, 0.1],
        power=0.85,
        lr=0.001,
        checkpoints_dir: str = ".",
        provenance_verbose: bool = False,
    ):
        super().__init__()
        self.save_hyperparameters()
        self.automatic_optimization = False

        self.latent_size = latent_size
        self.loss_weights = loss_weights
        self.lr = lr
        self.power = power
        self.checkpoints_dir = checkpoints_dir
        self.provenance_verbose = provenance_verbose
        os.makedirs(self.checkpoints_dir, exist_ok=True)

        self.generator = Generator(self.latent_size)
        self.discriminator = Discriminator(self.power)

        self.epoch_gen_loss = []
        self.epoch_disc_loss = []
        self.disc_epoch_test_loss = []
        self.gen_epoch_test_loss = []
        self.index = 0
        self.train_history = defaultdict(list)
        self.test_history = defaultdict(list)
        # self.pklfile = checkpoint_path
        # checkpoint_dir = os.path.dirname(checkpoint_path)
        # os.makedirs(checkpoint_dir, exist_ok=True)

    @property
    def itwinai_logger(self) -> BaseItwinaiLogger:
        try:
            itwinai_logger = self.trainer.itwinai_logger
        except AttributeError:
            print("WARNING: itwinai_logger attribute not set " f"in {self.__class__.__name__}")
            itwinai_logger = None
        return itwinai_logger

    def on_fit_start(self) -> None:
        if self.itwinai_logger:
            # Log hyper-parameters
            self.itwinai_logger.save_hyperparameters(self.hparams)

    def BitFlip(self, x, prob=0.05):
        """
        Flips a single bit according to a certain probability.

        Args:
            x (list): list of bits to be flipped
            prob (float): probability of flipping one bit

        Returns:
            list: List of flipped bits

        """
        x = np.array(x)
        selection = np.random.uniform(0, 1, x.shape) < prob
        x[selection] = 1 * np.logical_not(x[selection])
        return x

    def mean_absolute_percentage_error(self, y_true, y_pred):
        return torch.mean(torch.abs((y_true - y_pred) / (y_true + 1e-7))) * 100

    def compute_global_loss(self, labels, predictions, loss_weights=(3, 0.1, 25, 0.1)):
        # Can be initialized outside
        binary_crossentropy_object = nn.BCEWithLogitsLoss(reduction="none")
        # there is no equivalent in pytorch for
        # tf.keras.losses.MeanAbsolutePercentageError --> using the
        # custom "mean_absolute_percentage_error" above!
        mean_absolute_percentage_error_object1 = self.mean_absolute_percentage_error(
            predictions[1], labels[1]
        )
        mean_absolute_percentage_error_object2 = self.mean_absolute_percentage_error(
            predictions[3], labels[3]
        )
        mae_object = nn.L1Loss(reduction="none")

        binary_example_loss = (
            binary_crossentropy_object(predictions[0], labels[0]) * loss_weights[0]
        )

        # mean_example_loss_1 = mean_absolute_percentage_error_object(
        # predictions[1], labels[1]) * loss_weights[1]
        mean_example_loss_1 = mean_absolute_percentage_error_object1 * loss_weights[1]

        mae_example_loss = mae_object(predictions[2], labels[2]) * loss_weights[2]

        # mean_example_loss_2 = mean_absolute_percentage_error_object(
        # predictions[3], labels[3]) * loss_weights[3]
        mean_example_loss_2 = mean_absolute_percentage_error_object2 * loss_weights[3]

        binary_loss = binary_example_loss.mean()
        mean_loss_1 = mean_example_loss_1.mean()
        mae_loss = mae_example_loss.mean()
        mean_loss_2 = mean_example_loss_2.mean()

        return [binary_loss, mean_loss_1, mae_loss, mean_loss_2]

    def forward(self, z):
        return self.generator(z)

    def training_step(self, batch, batch_idx):
        image_batch, energy_batch, ang_batch, ecal_batch = (
            batch["X"],
            batch["Y"],
            batch["ang"],
            batch["ecal"],
        )

        image_batch = image_batch.permute(0, 4, 1, 2, 3)

        image_batch = image_batch.to(self.device)
        energy_batch = energy_batch.to(self.device)
        ang_batch = ang_batch.to(self.device)
        ecal_batch = ecal_batch.to(self.device)

        optimizer_discriminator, optimizer_generator = self.optimizers()
        batch_size = energy_batch.shape[0]

        noise = torch.randn(
            (batch_size, self.latent_size - 2), dtype=torch.float32, device=self.device
        )
        # print(f'Energy elements: {energy_batch.numel} {energy_batch.shape}')
        # print(f'Angle elements: {ang_batch.numel} {ang_batch.shape}')
        generator_ip = torch.cat(
            (energy_batch.view(-1, 1), ang_batch.view(-1, 1), noise), dim=1
        )
        generated_images = self.generator(generator_ip)

        # Train discriminator first on real batch
        fake_batch = self.BitFlip(np.ones(batch_size).astype(np.float32))
        fake_batch = torch.tensor([[el] for el in fake_batch]).to(self.device)
        labels = [fake_batch, energy_batch, ang_batch, ecal_batch]

        predictions = self.discriminator(image_batch)
        # print("calculating real_batch_loss...")
        real_batch_loss = self.compute_global_loss(labels, predictions, self.loss_weights)
        if self.itwinai_logger:
            self.itwinai_logger.log(
                item=sum(real_batch_loss),
                identifier="real_batch_loss",
                kind="metric",
                step=self.global_step,
                batch_idx=batch_idx,
                context="training",
            )

        # self.log("real_batch_loss", sum(real_batch_loss),
        #          prog_bar=True, on_step=True, on_epoch=True, sync_dist=True)
        # print("real batch disc train")
        # the following 3 lines correspond in tf version to:
        # gradients = tape.gradient(real_batch_loss,
        # discriminator.trainable_variables)
        # optimizer_discriminator.apply_gradients(zip(gradients,
        #  discriminator.trainable_variables)) in Tensorflow
        optimizer_discriminator.zero_grad()
        self.manual_backward(sum(real_batch_loss))
        # sum(real_batch_loss).backward()
        # real_batch_loss.backward()
        optimizer_discriminator.step()

        # Train discriminator on the fake batch
        fake_batch = self.BitFlip(np.zeros(batch_size).astype(np.float32))
        fake_batch = torch.tensor([[el] for el in fake_batch]).to(self.device)
        labels = [fake_batch, energy_batch, ang_batch, ecal_batch]

        predictions = self.discriminator(generated_images)

        fake_batch_loss = self.compute_global_loss(labels, predictions, self.loss_weights)
        # self.log("fake_batch_loss", sum(fake_batch_loss),
        #          prog_bar=True, on_step=True, on_epoch=True, sync_dist=True)

        if self.itwinai_logger:
            self.itwinai_logger.log(
                item=sum(fake_batch_loss),
                identifier="fake_batch_loss",
                kind="metric",
                step=self.global_step,
                batch_idx=batch_idx,
                context="training",
            )

        # print("fake batch disc train")
        # the following 3 lines correspond to
        # gradients = tape.gradient(fake_batch_loss,
        # discriminator.trainable_variables)
        # optimizer_discriminator.apply_gradients(zip(gradients,
        # discriminator.trainable_variables)) in Tensorflow
        optimizer_discriminator.zero_grad()
        self.manual_backward(sum(fake_batch_loss))
        # sum(fake_batch_loss).backward()
        optimizer_discriminator.step()

        # avg_disc_loss = (sum(real_batch_loss) + sum(fake_batch_loss)) / 2

        trick = np.ones(batch_size).astype(np.float32)
        fake_batch = torch.tensor([[el] for el in trick]).to(self.device)
        labels = [fake_batch, energy_batch.view(-1, 1), ang_batch, ecal_batch]

        gen_losses_train = []
        # Train generator twice using combined model
        for _ in range(2):
            noise = torch.randn((batch_size, self.latent_size - 2)).to(self.device)
            generator_ip = torch.cat(
                (energy_batch.view(-1, 1), ang_batch.view(-1, 1), noise), dim=1
            )

            generated_images = self.generator(generator_ip)
            predictions = self.discriminator(generated_images)

            loss = self.compute_global_loss(labels, predictions, self.loss_weights)
            # self.log("gen_loss", sum(loss), prog_bar=True,
            #          on_step=True, on_epoch=True, sync_dist=True)

            if self.itwinai_logger:
                self.itwinai_logger.log(
                    item=sum(loss),
                    identifier="gen_loss",
                    kind="metric",
                    step=self.global_step,
                    batch_idx=batch_idx,
                    context="training",
                )

            # print("gen train")
            optimizer_generator.zero_grad()
            self.manual_backward(sum(loss))
            # sum(loss).backward()
            optimizer_generator.step()

            for el in loss:
                gen_losses_train.append(el)

        avg_generator_loss = sum(gen_losses_train) / len(gen_losses_train)
        # self.log("generator_loss", avg_generator_loss.item(),
        #          prog_bar=True, on_step=True, on_epoch=True, sync_dist=True)

        if self.itwinai_logger:
            self.itwinai_logger.log(
                item=avg_generator_loss.item(),
                identifier="generator_loss",
                kind="metric",
                step=self.global_step,
                batch_idx=batch_idx,
                context="training",
            )
            # Log provenance information
            if self.provenance_verbose:
                # Log provenance at every training step
                self._log_provenance(context="training")

        # avg_generator_loss = [(a + b) / 2 for a, b in zip(*gen_losses_train)]
        # self.log("generator_loss", sum(avg_generator_loss), prog_bar=True,
        # on_step=True, on_epoch=True, sync_dist=True)

        gen_losses = []
        # I'm not returning anything as in pl you do not return anything when
        # you back-propagate manually
        # return_loss = real_batch_loss
        real_batch_loss = [
            real_batch_loss[0],
            real_batch_loss[1],
            real_batch_loss[2],
            real_batch_loss[3],
        ]
        fake_batch_loss = [
            fake_batch_loss[0],
            fake_batch_loss[1],
            fake_batch_loss[2],
            fake_batch_loss[3],
        ]
        gen_batch_loss = [
            gen_losses_train[0],
            gen_losses_train[1],
            gen_losses_train[2],
            gen_losses_train[3],
        ]
        gen_losses.append(gen_batch_loss)
        gen_batch_loss = [
            gen_losses_train[4],
            gen_losses_train[5],
            gen_losses_train[6],
            gen_losses_train[7],
        ]
        gen_losses.append(gen_batch_loss)

        real_batch_loss = [el.cpu().detach().numpy() for el in real_batch_loss]
        real_batch_loss_total_loss = np.sum(real_batch_loss)
        new_real_batch_loss = [real_batch_loss_total_loss]
        for i_weights in range(len(real_batch_loss)):
            new_real_batch_loss.append(
                real_batch_loss[i_weights] / self.loss_weights[i_weights]
            )
        real_batch_loss = new_real_batch_loss

        fake_batch_loss = [el.cpu().detach().numpy() for el in fake_batch_loss]
        fake_batch_loss_total_loss = np.sum(fake_batch_loss)
        new_fake_batch_loss = [fake_batch_loss_total_loss]
        for i_weights in range(len(fake_batch_loss)):
            new_fake_batch_loss.append(
                fake_batch_loss[i_weights] / self.loss_weights[i_weights]
            )
        fake_batch_loss = new_fake_batch_loss

        # if ecal sum has 100% loss(generating empty events) then end
        # the training
        if fake_batch_loss[3] == 100.0 and self.index > 10:
            # print("Empty image with Ecal loss equal to 100.0 "
            #       f"for {self.index} batch")
            torch.save(
                self.generator.state_dict(),
                os.path.join(self.checkpoints_dir, "generator_weights.pth"),
            )
            torch.save(
                self.discriminator.state_dict(),
                os.path.join(self.checkpoints_dir, "discriminator_weights.pth"),
            )
            if self.itwinai_logger:
                self.itwinai_logger.log(
                    item=os.path.join(self.checkpoints_dir, "generator_weights.pth"),
                    identifier="final_generator_weights",
                    kind="artifact",
                    context="training",
                )
                self.itwinai_logger.log(
                    item=os.path.join(self.checkpoints_dir, "discriminator_weights.pth"),
                    identifier="final_discriminator_weights",
                    kind="artifact",
                    context="training",
                )

            # print("real_batch_loss", real_batch_loss)
            # print("fake_batch_loss", fake_batch_loss)
            sys.exit()

        # append mean of discriminator loss for real and fake events
        self.epoch_disc_loss.append(
            [(a + b) / 2 for a, b in zip(real_batch_loss, fake_batch_loss)]
        )

        gen_losses[0] = [el.cpu().detach().numpy() for el in gen_losses[0]]
        gen_losses_total_loss = np.sum(gen_losses[0])
        new_gen_losses = [gen_losses_total_loss]
        for i_weights in range(len(gen_losses[0])):
            new_gen_losses.append(gen_losses[0][i_weights] / self.loss_weights[i_weights])
        gen_losses[0] = new_gen_losses

        gen_losses[1] = [el.cpu().detach().numpy() for el in gen_losses[1]]
        gen_losses_total_loss = np.sum(gen_losses[1])
        new_gen_losses = [gen_losses_total_loss]
        for i_weights in range(len(gen_losses[1])):
            new_gen_losses.append(gen_losses[1][i_weights] / self.loss_weights[i_weights])
        gen_losses[1] = new_gen_losses

        generator_loss = [(a + b) / 2 for a, b in zip(*gen_losses)]

        self.epoch_gen_loss.append(generator_loss)

        # # MB: verify weight synchronization among workers
        # # Ref: https://github.com/Lightning-AI/lightning/issues/9237
        # disc_w = self.discriminator.conv1.weight.reshape(-1)[0:5]
        # gen_w = self.generator.conv1.weight.reshape(-1)[0:5]
        # print(f"DISC w: {disc_w}")
        # print(f"GEN w: {gen_w}")

        # self.index += 1 #this might be moved after test cycle

        # logging of gen and disc loss done by Trainer
        # self.log('epoch_gen_loss', self.epoch_gen_loss, on_step=True,
        #  on_epoch=True, sync_dist=True)
        # self.log('epoch_disc_loss', self.epoch_disc_loss, on_step=True, o
        # n_epoch=True, sync_dist=True)

        # return avg_disc_loss + avg_generator_loss

    def on_train_epoch_end(self):
        if not self.provenance_verbose:
            # Log provenance only at the end of an epoch
            self._log_provenance(context="training")

        discriminator_train_loss = np.mean(np.array(self.epoch_disc_loss), axis=0)
        generator_train_loss = np.mean(np.array(self.epoch_gen_loss), axis=0)

        self.train_history["generator"].append(generator_train_loss)
        self.train_history["discriminator"].append(discriminator_train_loss)

        print("-" * 65)
        ROW_FMT = "{0:<20s} | {1:<4.2f} | {2:<10.2f} | " "{3:<10.2f}| {4:<10.2f} | {5:<10.2f}"
        print(ROW_FMT.format("generator (train)", *self.train_history["generator"][-1]))
        print(
            ROW_FMT.format("discriminator (train)", *self.train_history["discriminator"][-1])
        )

        torch.save(
            self.generator.state_dict(),
            os.path.join(self.checkpoints_dir, "generator_weights.pth"),
        )
        torch.save(
            self.discriminator.state_dict(),
            os.path.join(self.checkpoints_dir, "discriminator_weights.pth"),
        )

        if self.itwinai_logger:
            self.itwinai_logger.log(
                item=os.path.join(self.checkpoints_dir, "generator_weights.pth"),
                identifier="ckpts/generator_weights_epoch_" + str(self.current_epoch),
                kind="artifact",
                context="training",
            )
            self.itwinai_logger.log(
                item=self.generator,
                identifier="generator_epoch_" + str(self.current_epoch),
                kind="model",
                context="training",
            )
            self.itwinai_logger.log(
                item=os.path.join(self.checkpoints_dir, "discriminator_weights.pth"),
                identifier="ckpts/discriminator_weights_epoch_" + str(self.current_epoch),
                kind="artifact",
                context="training",
            )

        # with open(self.pklfile, "wb") as f:
        #     pickle.dump({"train": self.train_history,
        #                 "test": self.test_history}, f)

        # pickle.dump({"train": self.train_history}, open(self.pklfile, "wb"))
        print("train-loss:" + str(self.train_history["generator"][-1][0]))

    def validation_step(self, batch, batch_idx):
        image_batch, energy_batch, ang_batch, ecal_batch = (
            batch["X"],
            batch["Y"],
            batch["ang"],
            batch["ecal"],
        )

        image_batch = image_batch.permute(0, 4, 1, 2, 3)

        image_batch = image_batch.to(self.device)
        energy_batch = energy_batch.to(self.device)
        ang_batch = ang_batch.to(self.device)
        ecal_batch = ecal_batch.to(self.device)

        batch_size = energy_batch.shape[0]

        # Generate Fake events with same energy and angle as data batch
        noise = torch.randn(
            (batch_size, self.latent_size - 2), dtype=torch.float32, device=self.device
        )

        generator_ip = torch.cat(
            (energy_batch.view(-1, 1), ang_batch.view(-1, 1), noise), dim=1
        )
        generated_images = self.generator(generator_ip)

        # concatenate to fake and real batches
        X = torch.cat((image_batch, generated_images), dim=0)

        # y = np.array([1] * batch_size \
        # + [0] * batch_size).astype(np.float32)
        y = torch.tensor([1] * batch_size + [0] * batch_size, dtype=torch.float32).to(
            self.device
        )
        y = y.view(-1, 1)

        ang = torch.cat((ang_batch, ang_batch), dim=0)
        ecal = torch.cat((ecal_batch, ecal_batch), dim=0)
        aux_y = torch.cat((energy_batch, energy_batch), dim=0)

        # y = [[el] for el in y]
        labels = [y, aux_y, ang, ecal]

        # Calculate discriminator loss
        disc_eval = self.discriminator(X)
        disc_eval_loss = self.compute_global_loss(labels, disc_eval, self.loss_weights)

        # Calculate generator loss
        trick = np.ones(batch_size).astype(np.float32)
        fake_batch = torch.tensor([[el] for el in trick]).to(self.device)
        # fake_batch = [[el] for el in trick]
        labels = [fake_batch, energy_batch, ang_batch, ecal_batch]

        generated_images = self.generator(generator_ip)
        gen_eval = self.discriminator(generated_images)
        gen_eval_loss = self.compute_global_loss(labels, gen_eval, self.loss_weights)

        if self.itwinai_logger:
            self.itwinai_logger.log(
                item=sum(disc_eval_loss),
                identifier="val_discriminator_loss",
                kind="metric",
                step=self.global_step,
                batch_idx=batch_idx,
                context="validation",
            )
            self.itwinai_logger.log(
                item=sum(gen_eval_loss),
                identifier="val_generator_loss",
                kind="metric",
                step=self.global_step,
                batch_idx=batch_idx,
                context="validation",
            )
            # Log provenance information
            if self.provenance_verbose:
                # Log provenance at every validation step
                self._log_provenance(context="validation")

        # self.log('val_discriminator_loss', sum(
        #     disc_eval_loss), on_epoch=True, prog_bar=True, sync_dist=True)
        # self.log('val_generator_loss', sum(gen_eval_loss),
        #          on_epoch=True, prog_bar=True, sync_dist=True)

        disc_test_loss = [
            disc_eval_loss[0],
            disc_eval_loss[1],
            disc_eval_loss[2],
            disc_eval_loss[3],
        ]
        gen_test_loss = [
            gen_eval_loss[0],
            gen_eval_loss[1],
            gen_eval_loss[2],
            gen_eval_loss[3],
        ]

        # Configure the loss so it is equal to the original values
        disc_eval_loss = [el.cpu().detach().numpy() for el in disc_test_loss]
        disc_eval_loss_total_loss = np.sum(disc_eval_loss)
        new_disc_eval_loss = [disc_eval_loss_total_loss]
        for i_weights in range(len(disc_eval_loss)):
            new_disc_eval_loss.append(disc_eval_loss[i_weights] / self.loss_weights[i_weights])
        disc_eval_loss = new_disc_eval_loss

        gen_eval_loss = [el.cpu().detach().numpy() for el in gen_test_loss]
        gen_eval_loss_total_loss = np.sum(gen_eval_loss)
        new_gen_eval_loss = [gen_eval_loss_total_loss]
        for i_weights in range(len(gen_eval_loss)):
            new_gen_eval_loss.append(gen_eval_loss[i_weights] / self.loss_weights[i_weights])
        gen_eval_loss = new_gen_eval_loss

        self.index += 1
        # evaluate discriminator loss
        self.disc_epoch_test_loss.append(disc_eval_loss)
        # evaluate generator loss
        self.gen_epoch_test_loss.append(gen_eval_loss)

    def _log_provenance(self, context: str):
        if self.itwinai_logger:
            # Some provenance metrics
            self.itwinai_logger.log(
                item=self.current_epoch,
                identifier="epoch",
                kind="metric",
                step=self.current_epoch,
                context=context,
            )
            self.itwinai_logger.log(
                item=self,
                identifier=f"model_version_{self.current_epoch}",
                kind="model_version",
                step=self.current_epoch,
                context=context,
            )
            self.itwinai_logger.log(
                item=None,
                identifier=None,
                kind="system",
                step=self.current_epoch,
                context=context,
            )
            self.itwinai_logger.log(
                item=None,
                identifier=None,
                kind="carbon",
                step=self.current_epoch,
                context=context,
            )
            self.itwinai_logger.log(
                item=None,
                identifier="train_epoch_time",
                kind="execution_time",
                step=self.current_epoch,
                context=context,
            )

    def on_validation_epoch_end(self):
        if not self.provenance_verbose:
            # Log provenance only at the end of an epoch
            self._log_provenance(context="validation")

        discriminator_test_loss = np.mean(np.array(self.disc_epoch_test_loss), axis=0)
        generator_test_loss = np.mean(np.array(self.gen_epoch_test_loss), axis=0)

        self.test_history["generator"].append(generator_test_loss)
        self.test_history["discriminator"].append(discriminator_test_loss)

        print("-" * 65)
        ROW_FMT = "{0:<20s} | {1:<4.2f} | {2:<10.2f} | " "{3:<10.2f}| {4:<10.2f} | {5:<10.2f}"
        print(ROW_FMT.format("generator (test)", *self.test_history["generator"][-1]))
        print(ROW_FMT.format("discriminator (test)", *self.test_history["discriminator"][-1]))

        # # save loss dict to pkl file
        # with open(self.pklfile, "wb") as f:
        #     pickle.dump({"train": self.train_history,
        #                 "test": self.test_history}, f)
        # pickle.dump({"test": self.test_history}, open(self.pklfile, "wb"))
        # print("train-loss:" + str(self.train_history["generator"][-1][0]))

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        energy_batch, ang_batch = batch["Y"], batch["ang"]

        energy_batch = energy_batch.to(self.device)
        ang_batch = ang_batch.to(self.device)

        # Generate Fake events with same energy and angle as data batch
        noise = torch.randn(
            (energy_batch.shape[0], self.latent_size - 2),
            dtype=torch.float32,
            device=self.device,
        )

        # print(f"Reshape energy: {energy_batch.view(-1, 1).shape}")
        # print(f"Reshape angle: {ang_batch.view(-1, 1).shape}")
        # print(f"Noise: {noise.shape}")

        generator_ip = torch.cat(
            [energy_batch.view(-1, 1), ang_batch.view(-1, 1), noise], dim=1
        )
        # print(f"Generator input: {generator_ip.shape}")
        generated_images = self.generator(generator_ip)
        # print(f"Generated batch size {generated_images.shape}")
        return {"images": generated_images, "energies": energy_batch, "angles": ang_batch}

    def configure_optimizers(self):
        lr = self.lr

        optimizer_discriminator = torch.optim.RMSprop(self.discriminator.parameters(), lr)
        optimizer_generator = torch.optim.RMSprop(self.generator.parameters(), lr)

        if self.itwinai_logger:
            self.itwinai_logger.log(
                optimizer_discriminator, "optimizer_discriminator", kind="torch"
            )
            self.itwinai_logger.log(optimizer_generator, "optimizer_generator", kind="torch")

        return [optimizer_discriminator, optimizer_generator], []



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from pathlib import Path
from typing import List

import pandas as pd

from itwinai.scalability_report.data import read_scalability_metrics_from_csv
from itwinai.scalability_report.plot import (
    absolute_avg_epoch_time_plot,
    computation_fraction_bar_plot,
    gpu_bar_plot,
    relative_epoch_time_speedup_plot,
)
from itwinai.scalability_report.utils import (
    calculate_gpu_statistics,
    get_computation_fraction_data,
)


def epoch_time_report(
    log_dirs: List[Path] | List[str],
    plot_dir: Path | str,
    backup_dir: Path,
    do_backup: bool = False,
    plot_file_suffix: str = ".png",
) -> str | None:
    """Generates reports and plots for epoch training times across distributed training
    strategies, including a log-log plot of absolute average epoch times against the
    number of GPUs and a log-log plot of relative speedup as more GPUs are added. The
    function optionally creates backups of the data.

    Args:
        log_dirs (List[Path] | List[str]): List of paths to the directory containing CSV
            files with epoch time metrics. The files must include the columns "name", "nodes",
            "epoch_id", and "time".
        plot_dir (Path | str): Path to the directory where the generated plots will
            be saved.
        backup_dir (Path): Path to the directory where backups of the data will be stored
            if `do_backup` is True.
        do_backup (bool): Whether to create a backup of the epoch time data in the
            `backup_dir`. Defaults to False.
        plot_file_suffix (str): Suffix for the plot file names. Defaults to ".png".

    """
    if isinstance(plot_dir, str):
        plot_dir = Path(plot_dir)
    log_dir_paths = [Path(logdir) for logdir in log_dirs]

    epoch_time_expected_columns = {"name", "nodes", "epoch_id", "time"}

    # Reading data from all the logdirs and concatenating the results
    dataframes = []
    for log_dir in log_dir_paths:
        temp_df = read_scalability_metrics_from_csv(
            data_dir=log_dir, expected_columns=epoch_time_expected_columns
        )
        dataframes.append(temp_df)
    if not dataframes:
        return None
    epoch_time_df = pd.concat(dataframes)

    # Calculate the average time per epoch for each strategy and number of nodes
    print("\nAnalyzing Epoch Time Data...")
    avg_epoch_time_df = (
        epoch_time_df.groupby(["name", "nodes"])
        .agg(avg_epoch_time=("time", "mean"))
        .reset_index()
    )

    # Print the resulting table
    formatters = {"avg_epoch_time": "{:.2f} s".format}
    epoch_time_table = avg_epoch_time_df.to_string(index=False, formatters=formatters)

    # Create and save the figures
    absolute_fig, _ = absolute_avg_epoch_time_plot(avg_epoch_time_df=avg_epoch_time_df)
    relative_fig, _ = relative_epoch_time_speedup_plot(avg_epoch_time_df=avg_epoch_time_df)

    absolute_avg_time_plot_path = plot_dir / ("absolute_epoch_time" + plot_file_suffix)
    relative_speedup_plot_path = plot_dir / ("relative_epoch_time_speedup" + plot_file_suffix)

    absolute_fig.savefig(absolute_avg_time_plot_path)
    relative_fig.savefig(relative_speedup_plot_path)
    print(f"Saved absolute average time plot at '{absolute_avg_time_plot_path.resolve()}'.")
    print(f"Saved relative average time plot at '{relative_speedup_plot_path.resolve()}'.")

    if not do_backup:
        return epoch_time_table

    backup_dir.mkdir(exist_ok=True, parents=True)
    backup_path = backup_dir / "epoch_time_data.csv"
    epoch_time_df.to_csv(backup_path)
    print(f"Storing backup file at '{backup_path.resolve()}'.")
    return epoch_time_table


def gpu_data_report(
    log_dirs: List[Path] | List[str],
    plot_dir: Path | str,
    backup_dir: Path,
    do_backup: bool = False,
    plot_file_suffix: str = ".png",
) -> str | None:
    """Generates reports and plots for GPU energy consumption and utilization across
    distributed training strategies. Includes bar plots for energy consumption and GPU
    utilization by strategy and number of GPUs. The function optionally creates backups
    of the data.

    Args:
        log_dirs (List[Path] | List[str]): List of paths to the directory containing CSV
            files with GPU data. The files must include the columns "sample_idx",
            "utilization", "power", "local_rank", "node_idx", "num_global_gpus", "strategy",
            and "probing_interval".
        plot_dir (Path | str): Path to the directory where the generated plots will
            be saved.
        backup_dir (Path): Path to the directory where backups of the data will be stored
            if `do_backup` is True.
        do_backup (bool): Whether to create a backup of the GPU data in the `backup_dir`.
            Defaults to False.
    """
    if isinstance(plot_dir, str):
        plot_dir = Path(plot_dir)

    gpu_data_expected_columns = {
        "sample_idx",
        "utilization",
        "power",
        "local_rank",
        "node_idx",
        "num_global_gpus",
        "strategy",
        "probing_interval",
    }
    log_dir_paths = [Path(logdir) for logdir in log_dirs]
    dataframes = []
    for log_dir in log_dir_paths:
        temp_df = read_scalability_metrics_from_csv(
            data_dir=log_dir, expected_columns=gpu_data_expected_columns
        )
        dataframes.append(temp_df)
    if not dataframes:
        return None
    gpu_data_df = pd.concat(dataframes)

    print("\nAnalyzing Epoch Time Data...")
    gpu_data_statistics_df = calculate_gpu_statistics(
        gpu_data_df=gpu_data_df, expected_columns=gpu_data_expected_columns
    )
    formatters = {
        "total_energy_wh": "{:.2f} Wh".format,
        "utilization": "{:.2f} %".format,
    }
    gpu_data_table = gpu_data_statistics_df.to_string(index=False, formatters=formatters)

    energy_plot_path = plot_dir / ("gpu_energy_plot" + plot_file_suffix)
    utilization_plot_path = plot_dir / ("utilization_plot" + plot_file_suffix)
    energy_fig, _ = gpu_bar_plot(
        data_df=gpu_data_statistics_df,
        plot_title="Energy Consumption by Framework and Number of GPUs",
        y_label="Energy Consumption (Wh)",
        main_column="total_energy_wh",
    )
    utilization_fig, _ = gpu_bar_plot(
        data_df=gpu_data_statistics_df,
        plot_title="GPU Utilization by Framework and Number of GPUs",
        y_label="GPU Utilization (%)",
        main_column="utilization",
    )
    energy_fig.savefig(energy_plot_path)
    utilization_fig.savefig(utilization_plot_path)
    print(f"Saved GPU energy plot at '{energy_plot_path.resolve()}'.")
    print(f"Saved utilization plot at '{utilization_plot_path.resolve()}'.")

    if not do_backup:
        return gpu_data_table

    backup_dir.mkdir(exist_ok=True, parents=True)
    backup_path = backup_dir / "gpu_data.csv"
    gpu_data_df.to_csv(backup_path)
    print(f"Storing backup file at '{backup_path.resolve()}'.")
    return gpu_data_table


def communication_data_report(
    log_dirs: List[Path] | List[str],
    plot_dir: Path | str,
    backup_dir: Path,
    do_backup: bool = False,
    plot_file_suffix: str = ".png",
) -> str | None:
    """Generates reports and plots for communication and computation fractions across
    distributed training strategies. Includes a bar plot showing the fraction of time
    spent on computation vs communication for each strategy and GPU count. The function
    optionally creates backups of the data.

    Args:
        log_dirs (List[Path] | List[str]): List of paths to the directory containing CSV
            files with communication data. The files must include the columns "strategy",
            "num_gpus", "global_rank", "name", and "self_cuda_time_total".
        plot_dir (Path | str): Path to the directory where the generated plot will
            be saved.
        backup_dir (Path): Path to the directory where backups of the data will be stored
            if `do_backup` is True.
        do_backup (bool): Whether to create a backup of the communication data in the
            `backup_dir`. Defaults to False.
    """
    if isinstance(plot_dir, str):
        plot_dir = Path(plot_dir)

    communication_data_expected_columns = {
        "strategy",
        "num_gpus",
        "global_rank",
        "name",
        "self_cuda_time_total",
    }
    log_dir_paths = [Path(logdir) for logdir in log_dirs]
    dataframes = []
    for log_dir in log_dir_paths:
        temp_df = read_scalability_metrics_from_csv(
            data_dir=log_dir, expected_columns=communication_data_expected_columns
        )
        dataframes.append(temp_df)
    if not dataframes:
        return None
    communication_data_df = pd.concat(dataframes)

    print("\nAnalyzing Communication Data...")
    computation_fraction_df = get_computation_fraction_data(communication_data_df)

    formatters = {"computation_fraction": lambda x: "{:.2f} %".format(x * 100)}
    communication_data_table = computation_fraction_df.to_string(
        index=False, formatters=formatters
    )

    computation_fraction_plot_path = plot_dir / (
        "computation_fraction_plot" + plot_file_suffix
    )
    computation_fraction_fig, _ = computation_fraction_bar_plot(computation_fraction_df)
    computation_fraction_fig.savefig(computation_fraction_plot_path)
    print(f"Saved computation fraction plot at '{computation_fraction_plot_path.resolve()}'.")

    if not do_backup:
        return communication_data_table

    backup_dir.mkdir(exist_ok=True, parents=True)
    backup_path = backup_dir / "communication_data.csv"
    communication_data_df.to_csv(backup_path)
    print(f"Storing backup file at '{backup_path.resolve()}'.")
    return communication_data_table



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import logging
import os

# Create a logger for your library
logger = logging.getLogger(__name__)

# Get log level from env variable (default to INFO)
log_level = os.getenv("ITWINAI_LOG_LEVEL", "INFO")

# Validate log level
valid_levels = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}
if log_level.upper() not in valid_levels:
    logger.warning(
        f"ITWINAI_LOG_LEVEL was set to '{log_level}', which is not recognized. "
        f"Supported types are {', '.join(valid_levels)}."
    )
    log_level = "INFO"  # Default to INFO if invalid

# Apply log level
logger.setLevel(getattr(logging, log_level.upper()))

# Prevent duplicate handlers
if not logger.hasHandlers():
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(levelname)s] %(name)s:%(lineno)d - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

# Prevent logs from bubbling up to the root logger
logger.propagate = False



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Scaling test of Microsoft Deepspeed on Imagenet using Resnet."""

import os
from pathlib import Path
from timeit import default_timer as timer

import deepspeed
import torch
import torch.distributed as dist
import torchvision
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from utils import get_parser, imagenet_dataset, train_epoch

from itwinai.loggers import EpochTimeTracker
from itwinai.torch.reproducibility import seed_worker, set_seed


def main():
    # Parse CLI args
    parser = get_parser()
    parser = deepspeed.add_config_arguments(parser)
    args = parser.parse_args()

    # Check resources availability
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    is_distributed = use_cuda and torch.cuda.device_count() > 0
    torch_prng = set_seed(args.rnd_seed, deterministic_cudnn=False)

    shuffle = args.shuff and args.rnd_seed is None
    persistent_workers = args.nworker > 1
    epoch_time_save_dir = Path(args.epoch_time_directory)

    train_dataset = imagenet_dataset(args.data_dir, subset_size=args.subset_size)
    train_sampler = None

    if is_distributed:
        deepspeed.init_distributed(dist_backend=args.backend)

        local_world_size = torch.cuda.device_count()
        global_rank = dist.get_rank()
        local_rank = dist.get_rank() % local_world_size
        pin_memory = True
        train_sampler = DistributedSampler(train_dataset, shuffle=shuffle)

        # To fix problems with deepspeed OpenMPI rank conflicting with local rank
        os.environ["OMPI_COMM_WORLD_LOCAL_RANK"] = os.environ.get("LOCAL_RANK", "")
    else:
        # Use a single worker (either on GPU or CPU)
        local_world_size = 1
        global_rank = 0
        local_rank = 0
        pin_memory = False

    train_loader = DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        sampler=train_sampler,
        num_workers=args.nworker,
        pin_memory=pin_memory,
        persistent_workers=persistent_workers,
        prefetch_factor=args.prefetch,
        generator=torch_prng,
        worker_init_fn=seed_worker,
    )
    device = torch.device(f"cuda:{local_rank}" if use_cuda else "cpu")

    # Create CNN model
    model = torchvision.models.resnet152()
    model = model.to(device)

    # Initializing deepspeed distributed strategy
    deepspeed_config = {
        "train_micro_batch_size_per_gpu": args.batch_size,  # redundant
        "optimizer": {
            "type": "SGD",
            "params": {"lr": args.lr, "momentum": args.momentum},
        },
        "fp16": {"enabled": False},
        "zero_optimization": False,
    }
    distrib_model, optimizer, _, _ = deepspeed.initialize(
        args=args,
        model=model,
        model_parameters=model.parameters(),
        config_params=deepspeed_config,
    )

    # Start training loop
    if global_rank == 0:
        num_nodes = os.environ.get("SLURM_NNODES", "1")
        save_path = epoch_time_save_dir / f"epochtime_deepspeed-bl_{num_nodes}.csv"
        epoch_time_logger = EpochTimeTracker(
            strategy_name="deepspeed-bl",
            save_path=save_path,
            num_nodes=int(num_nodes),
        )

    start_time = timer()
    start_epoch = 1
    for epoch_idx in range(start_epoch, args.epochs + 1):
        epoch_start_time = timer()
        if is_distributed:
            train_sampler.set_epoch(epoch_idx)

        # Training
        train_epoch(
            model=distrib_model,
            device=device,
            train_loader=train_loader,
            optimizer=optimizer,
        )

        if global_rank == 0:
            epoch_elapsed_time = timer() - epoch_start_time
            epoch_time_logger.add_epoch_time(epoch_idx, epoch_elapsed_time)
            print(f"[{epoch_idx}/{args.epochs}] - time: {epoch_elapsed_time:.2f}s")

    if global_rank == 0:
        total_time = timer() - start_time
        print(f"Training finished - took {total_time:.2f}s")
        epoch_time_logger.save()

    # Clean-up
    if is_distributed:
        dist.barrier()
        deepspeed.sys.exit()


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Test distributed training strategies."""

from pathlib import Path
from typing import Any, Literal
from unittest.mock import patch

import pytest
import torch
import torch.nn as nn
from ray.train import RunConfig
from torch.nn import Linear
from torch.optim import SGD, Optimizer
from torch.utils.data import DataLoader, Dataset, DistributedSampler

from itwinai.distributed import get_adaptive_ray_scaling_config, ray_cluster_is_running
from itwinai.torch.distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    RayDDPStrategy,
    RayDeepSpeedStrategy,
    RayHorovodStrategy,
    RayTorchDistributedStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
)
from itwinai.torch.type import DistributedStrategyError, UninitializedStrategyError


class DummyDataset(Dataset):
    def __init__(self, size=100):
        self.size = size
        self.data = torch.randn(size, 10)
        self.targets = torch.randint(0, 2, (size,))

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.data[idx], self.targets[idx]


class BaseTestDistributedStrategy:
    @pytest.fixture(scope="module")
    def strategy(self):
        """To be overridden by subclasses with the specific strategy."""
        raise NotImplementedError

    @pytest.fixture
    def simple_model(self):
        return Linear(10, 2)

    @pytest.fixture
    def optimizer(self, simple_model: nn.Module):
        return SGD(simple_model.parameters(), lr=0.01)

    @pytest.fixture
    def dataset(self):
        return DummyDataset()

    def test_cluster_properties(self, strategy: TorchDistributedStrategy):
        """Test that ranks and other simple properties are computed correctly."""
        assert strategy.is_initialized
        assert isinstance(strategy.device(), str)
        assert isinstance(strategy.is_main_worker, bool)
        assert strategy.global_world_size() >= 1
        assert strategy.local_world_size() >= 1
        assert strategy.global_rank() >= 0
        assert strategy.local_rank() >= 0

    def test_init_exceptions(
        self, strategy: TorchDistributedStrategy, simple_model: nn.Module, optimizer: Any
    ):
        """Check that the init method cannot be called twice and that the other methods raise
        and exception if called when the strategy is not initialized."""
        # Test re-initialization
        with pytest.raises(DistributedStrategyError) as init_exc:
            strategy.init()
            assert "already initialized" in init_exc.value

        # Test initialized flag
        strategy.is_initialized = False

        with pytest.raises(UninitializedStrategyError):
            strategy.distributed(simple_model, optimizer)
        with pytest.raises(UninitializedStrategyError):
            strategy.global_rank()
        with pytest.raises(UninitializedStrategyError):
            strategy.local_rank()
        with pytest.raises(UninitializedStrategyError):
            strategy.local_world_size()
        with pytest.raises(UninitializedStrategyError):
            strategy.global_world_size()
        with pytest.raises(UninitializedStrategyError):
            strategy.device()
        with pytest.raises(UninitializedStrategyError):
            strategy.is_main_worker
        x = torch.ones(2)
        with pytest.raises(UninitializedStrategyError):
            strategy.gather(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.allgather_obj(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.gather_obj(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.clean_up()

        strategy.is_initialized = True

    def test_dataloader(self, strategy: TorchDistributedStrategy, dataset: Dataset):
        """Check that the dataloader is distributed correctly."""
        dataloader = strategy.create_dataloader(dataset, batch_size=16, shuffle=True)
        assert isinstance(dataloader, DataLoader)
        assert isinstance(dataloader.sampler, DistributedSampler)
        assert dataloader.sampler.rank == strategy.global_rank()
        assert dataloader.sampler.num_replicas == strategy.global_world_size()

    def test_gather_operations(self, strategy: TorchDistributedStrategy):
        """Test collective operations."""
        # Test tensor gather
        local_tensor = torch.tensor([strategy.global_rank()], device=strategy.device())
        gathered = strategy.gather(local_tensor)
        if strategy.is_main_worker:
            assert len(gathered) == strategy.global_world_size()
        else:
            assert gathered is None

        # Test tensor gather from CPU
        my_tensor = torch.ones(10) * strategy.global_rank()
        tensors = strategy.gather(my_tensor, dst_rank=0)
        if strategy.is_main_worker:
            assert torch.stack(tensors).sum() == sum(range(strategy.global_world_size())) * 10
        else:
            assert tensors is None

        # Test object gather
        local_obj = {"rank": strategy.global_rank()}
        gathered_obj = strategy.gather_obj(local_obj)
        if strategy.is_main_worker:
            assert len(gathered_obj) == strategy.global_world_size()
        else:
            assert gathered_obj is None

        # Test allgather
        all_gathered = strategy.allgather_obj(local_obj)
        assert len(all_gathered) == strategy.global_world_size()

    def test_ranks(self, strategy):
        """Check that the values of the ranks are as expected."""
        nnodes = strategy.global_world_size() // strategy.local_world_size()

        # Test local ranks
        lranks = strategy.gather_obj(strategy.local_rank(), dst_rank=0)
        if strategy.is_main_worker:
            assert len(lranks) == strategy.global_world_size()
            assert sum(lranks) == sum(range(strategy.local_world_size())) * nnodes
        else:
            assert lranks is None

        # Test global ranks
        granks = strategy.gather_obj(strategy.global_rank(), dst_rank=0)
        if strategy.is_main_worker:
            assert len(granks) == strategy.global_world_size()
            assert sum(granks) == sum(range(strategy.global_world_size()))
        else:
            assert granks is None


@pytest.mark.hpc
@pytest.mark.torch_dist
class TestTorchDDPStrategy(BaseTestDistributedStrategy):
    @pytest.fixture(scope="module")
    def strategy(self, ddp_strategy: TorchDistributedStrategy) -> TorchDDPStrategy:
        return ddp_strategy

    def test_init(self, strategy: TorchDDPStrategy):
        """Test specific initialization of TorchDDPStrategy."""
        assert strategy.backend in ["nccl", "gloo"]

        # Test initialization
        init_path = "torch.distributed.init_process_group"
        with patch(init_path, autospec=True) as mock_init_torch:
            strategy = TorchDDPStrategy(
                backend="nccl" if torch.cuda.is_available() else "gloo"
            )
            strategy.init()
            mock_init_torch.assert_called_once()

    def test_distributed_model(
        self, strategy: TorchDDPStrategy, simple_model: nn.Module, optimizer: Optimizer
    ):
        """Test NN model distribution."""
        from torch.nn.parallel import DistributedDataParallel

        dist_model, dist_optimizer, _ = strategy.distributed(simple_model, optimizer)
        assert isinstance(dist_model, DistributedDataParallel)
        assert hasattr(dist_model, "module") or isinstance(dist_model, nn.Module)
        assert isinstance(dist_optimizer, Optimizer)


@pytest.mark.hpc
@pytest.mark.deepspeed_dist
class TestDeepSpeedStrategy(BaseTestDistributedStrategy):
    @pytest.fixture(scope="module")
    def strategy(self, deepspeed_strategy: DeepSpeedStrategy) -> DeepSpeedStrategy:
        return deepspeed_strategy

    def test_init(self, strategy: DeepSpeedStrategy):
        """Test specific initialization of DeepSpeedStrategy."""
        assert strategy.backend in ["nccl", "gloo", "mpi"]
        assert hasattr(strategy, "deepspeed"), (
            "Lazy import of deepspeed not found in DeepSpeedStrategy class."
        )

        # Test initialization
        init_path = "deepspeed.init_distributed"
        with patch(init_path, autospec=True) as mock_init_ds:
            strategy = DeepSpeedStrategy(
                backend="nccl" if torch.cuda.is_available() else "gloo"
            )
            strategy.init()
            mock_init_ds.assert_called_once()

    def test_distributed_model(
        self, strategy: DeepSpeedStrategy, simple_model: nn.Module, optimizer: Optimizer
    ):
        """Test NN model distribution."""
        from deepspeed.runtime.engine import DeepSpeedEngine

        ds_config = {
            "train_batch_size": 16,
            "fp16": {"enabled": False},
            "optimizer": {"type": "SGD", "params": {"lr": 0.001}},
        }

        dist_model, dist_optimizer, _ = strategy.distributed(
            simple_model, optimizer=optimizer, config=ds_config
        )
        assert hasattr(dist_model, "module")
        assert isinstance(dist_model, DeepSpeedEngine)
        assert dist_optimizer is not None
        assert isinstance(dist_optimizer, Optimizer)


@pytest.mark.hpc
@pytest.mark.horovod_dist
class TestHorovodStrategy(BaseTestDistributedStrategy):
    @pytest.fixture(scope="module")
    def strategy(self, horovod_strategy: HorovodStrategy) -> HorovodStrategy:
        return horovod_strategy

    def test_init(self, strategy: HorovodStrategy):
        assert strategy.is_initialized
        assert hasattr(strategy, "hvd"), (
            "Lazy import of horovod not found in HorovodStrategy class."
        )

        # Test initialization
        init_path = "horovod.torch.init"
        with patch(init_path, autospec=True) as mock_init_ds:
            strategy = HorovodStrategy()
            strategy.init()
            mock_init_ds.assert_called_once()

    def test_distributed_model(self, strategy: HorovodStrategy, simple_model, optimizer):
        dist_model, dist_optimizer, _ = strategy.distributed(
            simple_model, optimizer=optimizer, op=strategy.hvd.Average
        )
        assert isinstance(dist_model, nn.Module)
        assert isinstance(dist_optimizer, Optimizer)
        assert hasattr(dist_optimizer, "synchronize"), (
            "synchronize() method not found for Horovod optimizer"
        )


@pytest.mark.hpc
@pytest.mark.ray_dist
@pytest.mark.parametrize(
    "strategy_name",
    [
        pytest.param("ddp"),
        pytest.param("deepspeed"),
        pytest.param("horovod"),
    ],
)
def test_ray_distributed_strategy(
    strategy_name: Literal["ddp"] | Literal["deepspeed"] | Literal["horovod"],
    shared_tmp_path: Path,
):
    import ray  # needed here

    assert ray_cluster_is_running(), "Ray cluster not detected. Aborting tests"

    # The worker function must be declared inline, ohterwise the Ray workers will not find it
    def ray_tests(config, strategy: RayTorchDistributedStrategy):
        """The tests below are a flattened version of BaseTestDistributedStrategy"""
        strategy.init()

        simple_model = Linear(10, 2)
        simple_optimizer = SGD(simple_model.parameters(), lr=0.01)

        # Strategy-specific tests
        if isinstance(strategy, RayDDPStrategy):
            # Test init
            assert hasattr(strategy, "ray_train"), (
                "Lazy import of ray.train not found in RayDDPStrategy class."
            )
            # Test distribute model
            from torch.nn.parallel import DistributedDataParallel

            dist_model, dist_optimizer, _ = strategy.distributed(
                simple_model, simple_optimizer
            )
            assert isinstance(dist_model, DistributedDataParallel)
            assert hasattr(dist_model, "module") or isinstance(dist_model, nn.Module)
            assert isinstance(dist_optimizer, Optimizer)

        elif isinstance(strategy, RayDeepSpeedStrategy):
            assert strategy.backend in ["nccl", "gloo", "mpi"]
            assert hasattr(strategy, "deepspeed"), (
                "Lazy import of deepspeed not found in RayDeepSpeedStrategy class."
            )

            # Test initialization
            init_path = "deepspeed.init_distributed"
            with patch(init_path, autospec=True) as mock_init_ds:
                strategy = RayDeepSpeedStrategy(
                    backend="nccl" if torch.cuda.is_available() else "gloo"
                )
                strategy.init()
                mock_init_ds.assert_called_once()

            # Test distribute model
            from deepspeed.runtime.engine import DeepSpeedEngine

            ds_config = {
                "train_batch_size": 16,
                "fp16": {"enabled": False},
                "optimizer": {"type": "SGD", "params": {"lr": 0.001}},
            }

            dist_model, dist_optimizer, _ = strategy.distributed(
                simple_model, optimizer=simple_optimizer, config=ds_config
            )
            assert hasattr(dist_model, "module")
            assert isinstance(dist_model, DeepSpeedEngine)
            assert dist_optimizer is not None
            assert isinstance(dist_optimizer, Optimizer)

        elif isinstance(strategy, RayHorovodStrategy):
            assert strategy.is_initialized
            assert hasattr(strategy, "hvd"), (
                "Lazy import of horovod not found in RayHorovodStrategy class."
            )

            # Test initialization
            init_path = "horovod.torch.init"
            with patch(init_path, autospec=True) as mock_init_ds:
                strategy = RayHorovodStrategy()
                strategy.init()
                mock_init_ds.assert_called_once()

            # Test distributed model
            dist_model, dist_optimizer, _ = strategy.distributed(
                simple_model, optimizer=simple_optimizer, op=strategy.hvd.Average
            )
            assert isinstance(dist_model, nn.Module)
            assert isinstance(dist_optimizer, Optimizer)
            assert hasattr(dist_optimizer, "synchronize"), (
                "synchronize() method not found for Horovod optimizer"
            )

        else:
            raise ValueError("Unrecognized strategy type")

        # Test cluster properties
        assert strategy.is_initialized
        assert isinstance(strategy.device(), str)
        assert isinstance(strategy.is_main_worker, bool)
        assert strategy.global_world_size() >= 1
        assert strategy.local_world_size() >= 1
        assert strategy.global_rank() >= 0
        assert strategy.local_rank() >= 0

        # Test re-initialization
        with pytest.raises(DistributedStrategyError) as init_exc:
            strategy.init()
            assert "already initialized" in init_exc.value

        # Test initialized flag
        strategy.is_initialized = False

        with pytest.raises(UninitializedStrategyError):
            strategy.distributed(simple_model, simple_optimizer)
        with pytest.raises(UninitializedStrategyError):
            strategy.global_rank()
        with pytest.raises(UninitializedStrategyError):
            strategy.local_rank()
        with pytest.raises(UninitializedStrategyError):
            strategy.local_world_size()
        with pytest.raises(UninitializedStrategyError):
            strategy.global_world_size()
        with pytest.raises(UninitializedStrategyError):
            strategy.device()
        with pytest.raises(UninitializedStrategyError):
            strategy.is_main_worker
        x = torch.ones(2)
        with pytest.raises(UninitializedStrategyError):
            strategy.gather(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.allgather_obj(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.gather_obj(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.clean_up()

        strategy.is_initialized = True

        # Test tensor gather
        local_tensor = torch.tensor([strategy.global_rank()], device=strategy.device())
        gathered = strategy.gather(local_tensor)
        if strategy.is_main_worker:
            assert len(gathered) == strategy.global_world_size()
        else:
            assert gathered is None

        # Test tensor gather from CPU
        my_tensor = torch.ones(10) * strategy.global_rank()
        tensors = strategy.gather(my_tensor, dst_rank=0)
        if strategy.is_main_worker:
            assert torch.stack(tensors).sum() == sum(range(strategy.global_world_size())) * 10
        else:
            assert tensors is None

        # Test object gather
        local_obj = {"rank": strategy.global_rank()}
        gathered_obj = strategy.gather_obj(local_obj)
        if strategy.is_main_worker:
            assert len(gathered_obj) == strategy.global_world_size()
        else:
            assert gathered_obj is None

        # Test allgather
        all_gathered = strategy.allgather_obj(local_obj)
        assert len(all_gathered) == strategy.global_world_size()

        # Check that the values of the ranks are as expected
        nnodes = strategy.global_world_size() // strategy.local_world_size()

        # Test local ranks
        lranks = strategy.gather_obj(strategy.local_rank(), dst_rank=0)
        if strategy.is_main_worker:
            assert len(lranks) == strategy.global_world_size()
            assert sum(lranks) == sum(range(strategy.local_world_size())) * nnodes
        else:
            assert lranks is None

        # Test global ranks
        granks = strategy.gather_obj(strategy.global_rank(), dst_rank=0)
        if strategy.is_main_worker:
            assert len(granks) == strategy.global_world_size()
            assert sum(granks) == sum(range(strategy.global_world_size()))
        else:
            assert granks is None

    # scaling_config = ScalingConfig(num_workers=2, use_gpu=False)
    scaling_config = get_adaptive_ray_scaling_config()
    run_config = RunConfig(storage_path=shared_tmp_path / "ray_checkpoints")

    match strategy_name:
        case "ddp":
            # This calls ray.init under the hood
            strategy = RayDDPStrategy()

            # Trainable
            test_function = ray.tune.with_parameters(ray_tests, strategy=strategy)

            # Create a trainer
            trainer = ray.train.torch.TorchTrainer(
                test_function,
                scaling_config=scaling_config,
                run_config=run_config,
            )
            trainer.fit()
        case "deepspeed":
            # This calls ray.init under the hood
            strategy = RayDeepSpeedStrategy(
                backend="nccl" if torch.cuda.is_available() else "gloo"
            )

            # Trainable
            test_function = ray.tune.with_parameters(ray_tests, strategy=strategy)

            # Create a trainer
            trainer = ray.train.torch.TorchTrainer(
                test_function,
                scaling_config=scaling_config,
                run_config=run_config,
            )
            trainer.fit()
        case "horovod":
            import ray.train.horovod

            # This calls ray.init under the hood
            strategy = RayHorovodStrategy()

            # Trainable
            test_function = ray.tune.with_parameters(ray_tests, strategy=strategy)

            # Create a trainer
            trainer = ray.train.horovod.HorovodTrainer(
                test_function,
                scaling_config=scaling_config,
                run_config=run_config,
            )
            trainer.fit()
        case _:
            raise ValueError("unrecognized strategy name")



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import copy
import re
from typing import Dict, List

# Template with default values
POD_TEMPLATE = {
    "apiVersion": "v1",
    "kind": "Pod",
    "metadata": {
        "name": "pod-template",
        "namespace": "default",
        "annotations": {
            "slurm-job.vk.io/flags": (
                "-p gpu --gres=gpu:1 --ntasks-per-node=1 --nodes=1 --time=00:10:00"
            ),
            "slurm-job.vk.io/pre-exec": "ls -la",
        },
    },
    "spec": {
        "containers": [
            {
                "name": "busy-container",
                "image": "/ceph/hpc/data/st2301-itwin-users/cern/hello-world-image.sif",
                "command": ["/bin/sh", "-c"],
                "args": ["sleep $(( 60 * 1 ))"],
                "imagePullPolicy": "Always",
                "resources": {
                    "limits": {"cpu": "48", "memory": "150Gi"},
                    "requests": {"cpu": "4", "memory": "20Gi"},
                },
            }
        ],
        "restartPolicy": "Always",
        "nodeSelector": {"kubernetes.io/hostname": "vega-new-vk"},
        "tolerations": [
            {"key": "virtual-node.interlink/no-schedule", "operator": "Exists"},
            {
                "effect": "NoExecute",
                "key": "node.kubernetes.io/not-ready",
                "operator": "Exists",
                "tolerationSeconds": 300,
            },
            {
                "effect": "NoExecute",
                "key": "node.kubernetes.io/unreachable",
                "operator": "Exists",
                "tolerationSeconds": 300,
            },
        ],
    },
}


def validate_pod_name(name: str) -> None:
    """Validates if a given string matches the Kubernetes pod naming convention.

    Args:
        name (str): The pod name to validate.

    Raises:
        ValueError: If the name does not match the required pattern.
    """
    pattern = r"[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*"
    if not re.fullmatch(pattern, name):
        raise ValueError(
            f"The name '{name}' is invalid for a Kubernetes pod. "
            "Pod names must match the pattern: "
            "'[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*'"
        )


def create_pod_manifest(
    name: str | None = None,
    annotations: Dict | None = None,
    container_name: str | None = None,
    image_path: str | None = None,
    cmd_args: List[str] | None = None,
    resources: Dict | None = None,
    target_node: str | None = None,
):
    """Creates a pod manifest with optional parameters to override default values
    for pod configuration.

    Args:
        name (str, optional): Name of the pod (metadata.name).
        annotations (Dict, optional): Annotations for the pod (metadata.annotations).
        container_name (str, optional): Name of the container (spec.containers[0].name).
        image_path (str, optional): Path to the container image
            (spec.containers[0].image).
        cmd_args (List[str], optional): Command arguments for the container
            (spec.containers[0].args).
        resources (Dict, optional): Resource requests and limits for the container
            (spec.containers[0].resources).
        target_node (str): target node of the node selector.

    Returns:
        Dict: A dictionary representing the pod manifest.
    """

    # Deep copy the template to avoid modifying the original
    pod_manifest = copy.deepcopy(POD_TEMPLATE)

    # Set metadata fields
    if name:
        validate_pod_name(name)
        pod_manifest["metadata"]["name"] = name
    if annotations:
        pod_manifest["metadata"]["annotations"].update(annotations)

    # Set container-specific fields
    container = pod_manifest["spec"]["containers"][0]
    if container_name:
        container["name"] = container_name
    if image_path:
        container["image"] = image_path
    if cmd_args:
        container["args"] = cmd_args
    if resources:
        container["resources"] = resources

    if target_node:
        pod_manifest["spec"]["nodeSelector"]["kubernetes.io/hostname"] = target_node

    return pod_manifest



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Downsample H5 files to a more manageable size."""

import h5py

IN_FILENAME = "large_file.h5"
OUT_FILENAME = "sample.h5"
MAXITEMS = 100

with h5py.File(IN_FILENAME, "r") as input_file:
    with h5py.File(OUT_FILENAME, "w") as outfile:
        for key in input_file.keys():
            print(input_file[key])
            shape = list(input_file[key].shape)
            shape[0] = MAXITEMS
            outfile.create_dataset_like(name=key, other=input_file[key], shape=tuple(shape))
            print(outfile[key])
            outfile[key][...] = input_file[key][:MAXITEMS]

        print("verify similarities")
        print(input_file["energy"][:10])
        print(outfile["energy"][:10])



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import importlib
from typing import List

from .exceptions import SanityCheckError

core_modules = [
    "itwinai",
    "itwinai.cli",
    "itwinai.components",
    "itwinai.distributed",
    "itwinai.loggers",
    "itwinai.parser",
    "itwinai.pipeline",
    "itwinai.serialization",
    "itwinai.type",
    "itwinai.utils",
    "itwinai.tests",
    "itwinai.tests.dummy_components",
    "itwinai.tests.exceptions",
    "itwinai.tests.sanity_check",
    "itwinai.plugins",
    "itwinai.scalability_report",
    "itwinai.scalability_report.data",
    "itwinai.scalability_report.utils",
    "itwinai.scalability_report.reports",
    "itwinai.scalability_report.plot",
]

torch_modules = [
    "itwinai.torch",
    "itwinai.torch.data",
    "itwinai.torch.gan",
    "itwinai.torch.models",
    "itwinai.torch.models.mnist",
    "itwinai.torch.config",
    "itwinai.torch.distributed",
    "itwinai.torch.inference",
    "itwinai.torch.mlflow",
    "itwinai.torch.reproducibility",
    "itwinai.torch.trainer",
    "itwinai.torch.type",
    "itwinai.torch.loggers",
    "itwinai.torch.tuning",
]

tensorflow_modules = [
    "itwinai.tensorflow",
    "itwinai.tensorflow.data",
    "itwinai.tensorflow.models",
    "itwinai.tensorflow.models.mnist",
    "itwinai.tensorflow.distributed",
    "itwinai.tensorflow.trainer",
    "itwinai.tensorflow.utils",
]


def run_sanity_check(modules: List[str]):
    """Run sanity checks by trying to import modules.

    Args:
        modules (List[str]): list of modules

    Raises:
        SanityCheckError: when some module cannot be imported.
    """
    failed_imports = []

    for module in modules:
        try:
            importlib.import_module(module)
            print(f"✅ Successfully imported: {module}")
        except ImportError as e:
            failed_imports.append((module, str(e)))
            print(f"❌ Failed to import: {module} - {e}")

    if failed_imports:
        err_msg = "\nSummary of failed imports:\n"
        for module, error in failed_imports:
            err_msg += f"Module: {module}, Error: {error}\n"

        raise SanityCheckError(
            "Not all itwinai modules could be successfully imported!\n" + err_msg
        )
    else:
        print("\nAll modules imported successfully!")


def sanity_check_slim():
    """Run sanity check on the installation
    of core modules of itwinai (neither itwinai.torch,
    nor itwinai.tensorflow)."""

    run_sanity_check(modules=core_modules)


def sanity_check_torch():
    """Run sanity check on the installation of itwinai
    for a torch environment."""
    run_sanity_check(modules=core_modules + torch_modules)


def sanity_check_tensorflow():
    """Run sanity check on the installation of itwinai
    for a tensorflow environment."""
    run_sanity_check(modules=core_modules + tensorflow_modules)


def sanity_check_all():
    """Run all sanity checks."""
    run_sanity_check(modules=core_modules + torch_modules + tensorflow_modules)
